import os
import yaml
import time
import numpy as np
import argparse
import traceback
from typing import Optional, List
import re

from sotodlib import core
import sotodlib.site_pipeline.util as sp_util
from sotodlib.preprocess import preprocess_util as pp_util
from sotodlib.preprocess import _Preprocess, Pipeline, processes

logger = sp_util.init_logger("preprocess")

def preprocess_obs(
    obs_id, 
    configs,  
    overwrite=False, 
    logger=None,
    obs_group=None
):
    """Meant to be run as part of a batched script, this function calls the
    preprocessing pipeline a specific Observation ID and saves the results in
    the ManifestDb specified in the configs.   

    Arguments
    ----------
    obs_id: string or ResultSet entry
        obs_id or obs entry that is passed to context.get_obs
    configs: string or dictionary
        config file or loaded config directory
    overwrite: bool
        if True, overwrite existing entries in ManifestDb
    logger: logging instance
        the logger to print to
    obs_group: list of strings
        List of obs_ids within group
    """

    if logger is None: 
        logger = sp_util.init_logger("preprocess")
    
    if type(configs) == str:
        configs = yaml.safe_load(open(configs, "r"))

    context = core.Context(configs["context_file"])

    source_list = configs.get('source_list', None)
    source_names = []
    for _s in source_list:
        if isinstance(_s, str):
            source_names.append(_s)
        elif len(_s) == 3:
            source_names.append(_s[0])
        else:
            raise ValueError('Invalid style of source')

    for i, source in enumerate(source_names):
        source_names[i] = re.sub('[^0-9a-zA-Z]+', '', source)
 
    if os.path.exists(configs['archive']['index']):
        logger.info(f"Mapping {configs['archive']['index']} for the "
                    "archive index.")
        db = core.metadata.ManifestDb(configs['archive']['index'])
    else:
        logger.info(f"Creating {configs['archive']['index']} for the "
                     "archive index.")
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('obs:obs_id')
        scheme.add_data_field('dataset')
        scheme.add_data_field('coverage')
        scheme.add_data_field('source_distance')
        if obs_group:
            scheme.add_data_field('obs_group')
        db = core.metadata.ManifestDb(
            configs['archive']['index'],
            scheme=scheme
        )

    pipe = Pipeline(configs["process_pipe"], plot_dir=configs["plot_dir"], logger=logger)

    logger.info(f"Beginning run for {obs_id}")

    aman = context.get_obs(obs_id, no_signal=True)
    proc_aman, success = pipe.run(aman)
    if success != 'end':
        return

    policy = pp_util.ArchivePolicy.from_params(configs['archive']['policy'])
    dest_file, dest_dataset = policy.get_dest(obs_id)
    logger.info(f"Saving data to {dest_file}:{dest_dataset}")
    proc_aman.save(dest_file, dest_dataset, overwrite=overwrite)

    # Update the index.
    db_data = {'obs:obs_id': obs_id,
                'dataset': dest_dataset}
    
    sso_footprint_process = False
    for process in configs['process_pipe']:
        if 'name' in process and 'sso_footprint' == process['name']:
            sso_footprint_process = True

    if sso_footprint_process:
        logger.info(f"Saving per source to database {db_data}")
        nearby_source_names = []
        for _source in proc_aman.sso_footprint._assignments.keys():
            nearby_source_names.append(_source)
        nearby_source_names_map = {key.casefold(): key for key in nearby_source_names}
        coverage = []
        distances = []
        for source_name in source_names:
            source_name_l = source_name.casefold()
            if source_name_l in nearby_source_names_map:
                source_name = nearby_source_names_map.get(source_name_l)
                for key in proc_aman.sso_footprint[source_name]._assignments.keys():
                    if 'ws' in key:
                        if proc_aman.sso_footprint[source_name][key]:
                            coverage.append(f"{source_name}:{key}")

                distances.append(f"{source_name}:{proc_aman.sso_footprint[source_name]['mean_distance']}")
        db_data['coverage'] = ','.join(coverage)
        db_data['source_distance'] = ','.join(distances)
    else:
        db_data['coverage'] = None
        db_data['source_distance'] = None

    if obs_group:
        db_data['obs_group'] = ','.join(obs_group)
    
    logger.info(f"Saving to database under {db_data}")
    if len(db.inspect(db_data)) == 0:
        h5_path = os.path.relpath(dest_file,
                start=os.path.dirname(configs['archive']['index']))
        db.add_entry(db_data, h5_path)

    if configs.get("lmsi_config", None) is not None:
        from pathlib import Path
        import lmsi.core as lmsi

        new_plots = os.path.join(configs["plot_dir"],
                                 f'{str(aman.timestamps[0])[:5]}',
                                 aman.obs_info.obs_id)
        if os.path.exists(new_plots):
            lmsi.core([Path(x.name) for x in Path(new_plots).glob("*.png")],
                      Path(configs["lmsi_config"]),
                      Path(os.path.join(new_plots, 'index.html')))

def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('configs', help="Preprocessing Configuration File")
    parser.add_argument(
        '--query', 
        help="Query to pass to the observation list. Use \\'string\\' to "
             "pass in strings within the query.",  
        type=str
    )
    parser.add_argument(
        '--obs-id',
        help="obs-id of particular observation if we want to run on just one"
    )
    parser.add_argument(
        '--overwrite',
        help="If true, overwrites existing entries in the database",
        action='store_true',
    )
    parser.add_argument(
        '--min-ctime',
        help="Minimum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--max-ctime',
        help="Maximum timestamp for the beginning of an observation list",
    )
    parser.add_argument(
        '--update-delay',
        help="Number of days (unit is days) in the past to start observation list.",
        type=int
    )
    parser.add_argument(
        '--tags',
        help="Observation tags. Ex: --tags 'jupiter' 'setting'",
        nargs='*',
        type=str
    )
    parser.add_argument(
        '--planet-obs',
        help="If true, takes all planet tags as logical OR and adjusts related configs",
        action='store_true',
    )
    parser.add_argument(
        '--verbosity',
        help="increase output verbosity. 0:Error, 1:Warning, 2:Info(default), 3:Debug",
        default=2,
        type=int
    )
    parser.add_argument(
        '--lat',
        help="If true, filter obs list to only keep the first obs_ids of those with \
              timestamps within 10 seconds, since lat obs_ids are split by optics tube. \
              We only need one for preprocess_obs loads the entire aman without signal.",
        action='store_true',
    )
    return parser

def main(
    configs: str,
    query: Optional[str] = None, 
    obs_id: Optional[str] = None, 
    overwrite: bool = False,
    min_ctime: Optional[int] = None,
    max_ctime: Optional[int] = None,
    update_delay: Optional[int] = None,
    tags: Optional[List[str]] = None,
    planet_obs: bool = False,
    verbosity: Optional[int] = None,
    lat: bool = False,
 ):
    configs, context = pp_util.get_preprocess_context(configs)
    logger = sp_util.init_logger("preprocess", verbosity=verbosity)
    
    obs_list = sp_util.get_obslist(context, query=query, obs_id=obs_id, min_ctime=min_ctime, 
                                   max_ctime=max_ctime, update_delay=update_delay, tags=tags, 
                                   planet_obs=planet_obs)
    
    if len(obs_list)==0:
        logger.warning(f"No observations returned from query: {query}")
    run_list = []
    
    if lat:
        time_tolerance = 10
        kept = []
        seen_ctimes = []
        obs_groups = []
        current_group = []

        for s in obs_list:
            try:
                ctime_str = s['obs_id'].split('_')[1]
                ctime = int(ctime_str)
            except (IndexError, ValueError):
                continue

            is_close = any(abs(ctime - seen) <= time_tolerance for seen in seen_ctimes)

            if not is_close:
                if current_group:
                    obs_groups.append(current_group)
                current_group = [s['obs_id']]
                kept.append(s)
                seen_ctimes = [ctime]
            else:
                current_group.append(s['obs_id'])
                seen_ctimes.append(ctime)

        if current_group:
            obs_groups.append(current_group)

        obs_list = kept
    else:
        obs_groups = None

    if overwrite or not os.path.exists(configs['archive']['index']):
        #run on all if database doesn't exist
        run_list = [o for o in obs_list]
    else:
        mask = []
        db = core.metadata.ManifestDb(configs['archive']['index'])
        for obs in obs_list:
            x = db.inspect({'obs:obs_id': obs["obs_id"]})
            if x is None or len(x) == 0:
                run_list.append(obs)
                mask.append(True)
            else:
                mask.append(False)
        
        if obs_groups:
            obs_groups = [group for keep, group in zip(mask, obs_groups) if keep]

    logger.info(f"Beginning to run preprocessing on {len(run_list)} observations")
    for i, obs in enumerate(run_list):
        logger.info(f"Processing obs_id: {obs['obs_id']}")
        try:
            preprocess_obs(obs["obs_id"], configs, overwrite=overwrite, logger=logger, 
                           obs_group=None if obs_groups is None else obs_groups[i])
        except Exception as e:
            logger.info(f"{type(e)}: {e}")
            logger.info(''.join(traceback.format_tb(e.__traceback__)))
            logger.info(f'Skiping obs:{obs["obs_id"]} and moving to the next')
            continue
            

if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)



import yaml
import argparse

from sotodlib import core
import sotodlib.site_pipeline.util as sp_util
from sotodlib.qa import metrics as qa_metrics
from sotodlib.site_pipeline.monitor import Monitor
from sotodlib.site_pipeline.jobdb import JobManager

logger = sp_util.init_logger("qa_metrics")


def main(config):
    """ Update all metrics specified in the config.
    """
    if isinstance(config, str):
        with open(config, "r") as fh:
            config = yaml.safe_load(fh)

    monitor = Monitor.from_configs(config["monitor"])
    context = core.Context(config["context_file"])

    # get JobDB configuration
    jdb_max_retry = 5  # mark a job as failed after this number of tries
    if isinstance(config["jobdb"], str):  # support previous convention
        jdb = JobManager(sqlite_file=config["jobdb"])
    else:
        jdb = JobManager(sqlite_file=config["jobdb"]["db_file"])
        jdb_max_retry = config["jobdb"].get("max_retry", jdb_max_retry)

    # get a list of metrics from the config
    metrics = []
    for m in config["metrics"]:
        # try to get an instance of this metric
        name = m.pop("name")
        try:
            metric_class = getattr(qa_metrics, name)
        except AttributeError:
            raise Exception(f"No metric named {name} was found.")
        # Instantiate class with remaining elements as kwargs
        metrics.append(metric_class(context=context, monitor=monitor, **m))

    # get a list of obs_id to process and add to the JobDB
    jobs = []
    all_obs_id = set()
    for m in metrics:
        # observations that have not been recorded to Influx
        new_obs = m.get_new_obs()

        # check if failed jobs exist already for these
        jclass = f"{m._influx_meas}.{m._influx_field}"
        skipped_jobs = jdb.get_jobs(jclass=jclass, jstate=["done", "failed", "ignored"])
        skipped_obs = set([j.tags["obs_id"] for j in skipped_jobs])
        logger.debug(f"Skipping {len(skipped_obs & new_obs)} / {len(new_obs)} obs_id for metric {m._influx_meas}.{m._influx_field}")
        new_obs -= skipped_obs

        # create list of jobs to process
        new_jobs = {}
        open_jobs = jdb.get_jobs(jclass=jclass, jstate=["open"])
        open_obs = [j.tags["obs_id"] for j in open_jobs]
        for oid in new_obs:
            if oid in open_obs:
                job = open_jobs[open_obs.index(oid)]
            else:
                job = jdb.create_job(jclass, {"obs_id": oid})
            new_jobs[oid] = job

        jobs.append(new_jobs)
        all_obs_id |= new_obs

    logger.info(f"Found {len(all_obs_id)} obs_id with new metrics to record.")

    # convenience function for marking a job failure
    def _mark_failure(job_id):
        with jdb.locked(job_id) as job:
            job.mark_visited()
            if job.visit_count > jdb_max_retry:
                logger.error(f"Reached max retries for {oid}.")
                job.jstate = "failed"
                return True
        return False

    n = len(all_obs_id)
    n_fail = 0
    # process one obs_id at a time
    for i, oid in enumerate(all_obs_id):
        logger.info(f"Recording metrics for obs_id {oid} ({i+1}/{n})...")
        # load metadata once
        meta_fail = False
        try:
            # this can fail if crucial metadata is unavailable
            meta = context.get_meta(oid, ignore_missing=True)
            # check that obsdb info is available
            if len(meta.obs_info.keys()) < 2:
                logger.warning("This observation is missing obs_info. Skipping.")
                raise ValueError("Missing obs_info.")
        except Exception as e:
            logger.error(f"Failed to load metadata for {oid}.\n{type(e)}: {e}")
            meta_fail = True

        # iterate over metrics and record, or mark failure to load metadata
        for m, m_obs in zip(metrics, jobs):
            if oid in m_obs:
                if meta_fail:
                    if _mark_failure(m_obs[oid]):  # returns True if we reached max retries
                        n_fail += 1
                    continue
                try:
                    m.process_and_record(oid, meta)
                    with jdb.locked(m_obs[oid]) as job:
                        job.mark_visited()
                        job.jstate = "done"
                except Exception:
                    logger.error(
                        f"Processing metric {m._influx_meas}.{m._influx_field} failed.",
                        exc_info=True
                    )
                    if _mark_failure(m_obs[oid]):  # returns True if we reached max retries
                        n_fail += 1

    if n_fail > 0:
        raise RuntimeError(f"This run produced {n_fail} failures.")


def get_parser(parser=None):
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument('config', help="Metrics Configuration File")
    return parser


if __name__ == '__main__':
    sp_util.main_launcher(main, get_parser)



import numpy as np
from sotodlib import core
import logging

logger = logging.getLogger(__name__)


def default_model(telescope):
    """Returns default hwp_angle_model. This will be deleted in future.
    """

    model = core.AxisManager()
    if telescope == 'satp1':
        sign = core.AxisManager()
        sign.wrap('pid', 1)
        sign.wrap('offcenter', -1)
        model.wrap('mechanical_offset_1', np.deg2rad(-1.66 - 90 + 49.1))
        model.wrap('mechanical_offset_2', np.deg2rad(-1.66 + 90 + 49.1))

    elif telescope == 'satp2':
        sign = core.AxisManager()
        sign.wrap('pid', 1)
        sign.wrap('offcenter', -1)
        model.wrap('mechanical_offset_1', np.deg2rad(-1.66 - 90 + 7.7))
        model.wrap('mechanical_offset_2', np.deg2rad(-1.66 + 90 + 7.7))

    elif telescope == 'satp3':
        sign = core.AxisManager()
        sign.wrap('pid', -1)
        sign.wrap('offcenter', 1)
        model.wrap('mechanical_offset_1', np.deg2rad(-1.66 + 90 - 2.29))
        model.wrap('mechanical_offset_2', np.deg2rad(-1.66 - 90 - 2.29))

    else:
        raise ValueError('Not supported yet')

    model.wrap('sign_matrix', sign)
    model.wrap('sign_dependent_offset', np.deg2rad(-1. * 360 / 1140 * 3 / 2))
    model.wrap('model_version', 'v2')
    return model


def apply_hwp_angle_model(tod, on_sign_ambiguous='fail'):
    """
    Applies `hwp_angle_model` to the `hwp_solution` and construct hwp angle
    calibrated in the telescope frame.
    This will populate the calibrated hwp angle as `hwp_angle` attribute.
    `hwp_solution` is the hwp angle measured in the encoder frame.
    `hwp_angle_model` corrects the sign and offset of `hwp_solution`.

    Args:
        tod: AxisManager
        on_sign_ambiguous: Tolerance options for sign ambiguous
            fail: raise an error if there is any sign ambiguous
            pid: raise an error if pid sign does not exist
            offcenter: raise an error if offcenter sign does not exist

    Returns:
        tod: AxisManager with hwp_angle attribute
    """

    telescope = tod.obs_info.telescope
    if "lat" in telescope:
        return tod

    hwp = tod.get('hwp_solution', None)
    if hwp is None:
        raise ValueError('hwp_solution is missing')
    model = tod.get('hwp_angle_model', None)
    if model is None:
        logger.warn('hwp_angle_model metadata is missing. '
                    'Apply default model. This may be old.')
        model = default_model(telescope)

    assert model.model_version == 'v2'
    # construct sign
    methods = model.sign_matrix.keys()
    if on_sign_ambiguous in methods:
        sign = hwp[on_sign_ambiguous + '_direction']
        sign *= model.sign_matrix[on_sign_ambiguous]
        if sign == 0:
            raise ValueError('hwp rotation direction is ambiguous')
    elif on_sign_ambiguous == 'fail':
        available_signs = []
        for method in methods:
            _sign = hwp[method + '_direction']
            if _sign != 0:
                available_signs.append(_sign*model.sign_matrix[method])

        # check agreements of available estimation methods
        if len(available_signs) == 0:
            raise ValueError('hwp rotation direction is ambiguous')
        if np.all(np.array(available_signs) == available_signs[0]):
            sign = available_signs[0]
        else:
            raise ValueError('hwp rotation direction is ambiguous')
    else:
        raise ValueError('Invalid on_sign_ambiguous')

    # apply correction
    hwp_angle = np.mod(sign * (hwp.hwp_angle + model.sign_dependent_offset)
                + model[f'mechanical_offset_{hwp.primary_encoder}'], 2*np.pi)
    if 'hwp_angle' in tod._assignments.keys():
        tod.hwp_angle = hwp_angle
    else:
        tod.wrap('hwp_angle', hwp_angle, [(0, "samps")])

    return tod



"""Build Matching Database between readout_id and detector_id
"""

import os
import sys
import yaml
import h5py
import logging
import numpy as np
import datetime as dt
from argparse import ArgumentParser

from detmap.makemap import MapMaker
from sotodlib import core
from sotodlib.io.load_smurf import G3tSmurf, TuneSets
from sotodlib.io.metadata import write_dataset, read_dataset
from sotodlib.site_pipeline import util

logger = util.init_logger(__name__)

def parse_args():
    parser = ArgumentParser()
    parser.add_argument('-c', '--config-file', help=
    "Configuration File for running DetMap")
    parser.add_argument('--min-ctime', type=float, help=
    "Minimum ctime to search for TuneSets")
    parser.add_argument('--max-ctime', type=float, help=
    "Maximum ctime to search for TuneSets")
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('--debug', action='store_true')
    parser.add_argument('--log-file', help="Output log filename")
    args = parser.parse_args()
    return args


def main(args=None):
    if args is None:
        args = parse_args()
        
    if args.debug:
        logger.setLevel("DEBUG")
        
    if args.log_file is not None:
        formatter = util._ReltimeFormatter('%(asctime)s: %(message)s (%(levelname)s)')
        file_handler = logging.FileHandler(args.log_file)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    
    configs = yaml.safe_load(open(args.config_file, "r"))
    SMURF = G3tSmurf.from_configs(configs)
    session = SMURF.Session()
    
    if os.path.exists(configs["read_db"]):
        logger.info(f'Mapping {configs["read_db"]} for the archive index.')
        db = core.metadata.ManifestDb(configs["read_db"])
    else:
        logger.info(f'Creating {configs["read_db"]} for the archive index.')
        scheme = core.metadata.ManifestScheme()
        scheme.add_exact_match('dets:detset')
        scheme.add_data_field('dataset')
        db = core.metadata.ManifestDb(configs["read_db"], scheme=scheme)

    for array in configs["arrays"]:
        # Load Detector Information
        rs = read_dataset(configs["det_info"], array["name"])
        det_rs = core.metadata.merge_det_info(None, rs)
        det_info = core.metadata.loader.convert_det_info(det_rs, dets=det_rs["det_id"])
        det_info.restrict(
            "dets",
            det_info.dets.vals[det_info.wafer.array == array["name"]],
        )
        logger.info(
            f"Found {det_info.dets.count} detector_ids for Array {array['name']}"
        )
        # Find TuneSets for Each Array
        tunesets = session.query(TuneSets).filter(
            TuneSets.stream_id == array["stream_id"]
        )
        if args.min_ctime is not None:
            tunesets = tunesets.filter(
                TuneSets.start >= dt.datetime.utcfromtimestamp(args.min_ctime)
            )
        if args.max_ctime is not None:
            tunesets = tunesets.filter(
                TuneSets.start <= dt.datetime.utcfromtimestamp(args.max_ctime)
            )
        tunesets = tunesets.all()
        logger.info(f"Found {len(tunesets)} TuneSets for Array {array['name']}")
        # Run mapping on each TuneSet.
        for ts in tunesets:
            bgmap_path = None
            mapping = array["mapping"]
            dest_dataset = f"{ts.stream_id}_{ts.name}_mapping_v{mapping['version']}"
            if dest_dataset in list(db.get_entries(["dataset"])["dataset"]) and not args.overwrite:
                logger.debug(f"Dataset {dest_dataset} already exists")
                continue
            # Setup Mapping for Each Mapping Type
            map_maker = MapMaker(north_is_highband=array["north_is_highband"],
                                 array_name=array["name"],
                                 mapping_strategy=mapping["strategy"],
                                 **mapping["params"])
            logger.info(f"Making Map for {ts.path}")
            try:
                array_map = map_maker.make_map_smurf(tunefile=ts.path,
                                                     bgmap_path=bgmap_path)
            except:
                logger.warning(f"Map Maker Failed on {ts.path}")
                continue

            readout_ids, bands, channels = zip(*[
                (ch.name, ch.band, ch.channel) for ch in ts.channels
            ])
            bands = np.array(bands)
            channels = np.array(channels)

            rs_list = core.metadata.ResultSet(
                keys=["dets:det_id",
                      "dets:readout_id"]
            )
            # loop through detector IDs and find readout ID matches
            for tune in array_map.get_inst_output(add_back_null=False):
                det_id = tune.detector_id
                if det_id is None:
                    logger.warning('The DetMap issued "det_id" is should not allowed' +
                                   'to be None in the return from ' +
                                   'detmap.data_io.channel_assignment.OperateTuneData.get_inst_output()')
                    continue
                i_did = np.where(det_info.dets.vals == det_id)[0]
                if len(i_did) != 1:
                    logger.warning(f"Map returned detector_id, {det_id}, "
                                   "not in database, what happened?")
                    continue
                i_rid = np.where(np.all([
                    bands == tune.smurf_band,
                    channels == tune.smurf_channel], axis=0))[0]
                if len(i_rid) != 1:
                    logger.debug(f"Detector {det_id} not found in Tuneset {ts.name}")
                    """ If we want to track which detectors do not have readout_ids it
                        would go here, but has trouble with dets:readout_id loading in 
                        the current framework."""
                    continue
                i_rid = i_rid[0]
                rs_list.append({"dets:det_id": det_id,
                                "dets:readout_id": readout_ids[i_rid]})

            for rid in readout_ids:
                if rid in rs_list["dets:readout_id"]:
                    continue
                rs_list.append({"dets:det_id": "NO_MATCH",
                                "dets:readout_id": rid})

            write_dataset(rs_list, configs["read_info"], dest_dataset, args.overwrite)

            if dest_dataset not in list(db.get_entries(["dataset"])["dataset"]):
                # Update the index.
                db_data = {'dets:detset': ts.name,
                           'dataset': dest_dataset}
                db.add_entry(db_data, configs["read_info"])


if __name__ == "__main__":
    main()



import sqlite3
import gzip
import os


def sqlite_to_file(db, filename, overwrite=True, fmt=None):
    """Write an sqlite db to file.  Supports several output formats.

    Args:
      db (sqlite3.Connection): the sqlite3 database connection.
      filename (str): the path to the output file.
      overwrite (bool): whether an existing file should be
        overwritten.
      fmt (str): 'sqlite', 'dump', or 'gz'.  Defaults to 'sqlite'
        unless the filename ends with '.gz', in which case it is 'gz'.

    """
    if fmt is None:
        if filename.endswith('.gz'):
            fmt = 'gz'
        else:
            fmt = 'sqlite'
    if os.path.exists(filename) and not overwrite:
        raise RuntimeError(f'File {filename} exists; remove or pass '
                           'overwrite=True.')
    if fmt == 'sqlite':
        if os.path.exists(filename):
            os.remove(filename)
        new_db = sqlite3.connect(filename)
        script = ' '.join(db.iterdump())
        new_db.executescript(script)
        new_db.commit()
    elif fmt == 'dump':
        with open(filename, 'w') as fout:
            for line in db.iterdump():
                fout.write(line)
    elif fmt == 'gz':
        with gzip.GzipFile(filename, 'wb') as fout:
            for line in db.iterdump():
                fout.write(line.encode('utf-8'))
    else:
        raise RuntimeError(f'Unknown format "{fmt}" requested.')

def sqlite_from_file(filename, fmt=None, force_new_db=True):
    """Instantiate an sqlite3.Connection and return it, with the data
    copied in from the specified file. The function can either map the database
    file directly, or map a copy of the database in memory (see force_new_db
    parameter).

    Args:
      filename (str): path to the file.
      fmt (str): format of the input; see to_file for details.
      force_new_db (bool): Used if connecting to an sqlite database. If True the
        databas is copied into memory and if False returns a connection to the 
        database without reading it into memory

    """
    if fmt is None:
        fmt = 'sqlite'
        if filename.endswith('.gz'):
            fmt = 'gz'
    if fmt == 'sqlite':
        db0 = sqlite3.connect(f'file:{filename}?mode=ro', uri=True)
        if not force_new_db:
            return db0
        data = ' '.join(db0.iterdump())
    elif fmt == 'dump':
        with open(filename, 'r') as fin:
            data = fin.read()
    elif fmt == 'gz':
        with gzip.GzipFile(filename, 'r') as fin:
            data = fin.read().decode('utf-8')
    else:
        raise RuntimeError(f'Unknown format "{fmt}" requested.')
    db = sqlite3.connect(':memory:')
    db.executescript(data)
    return db




# Copyright (c) 2023-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import warnings

import astropy.units as u
import numpy as np

try:
    import pandas as pd
    from detmap.data_io.solution_select import AvailableSolutions
    available_solutions = AvailableSolutions()
    detmap_available = True
except:
    detmap_available = False

mappings = {}

# SAT1 MF wafers  = w25-w31
# SAT2 MF wafers  = w32-w38
# SAT3 HF wafers  = w06-w12
# SAT4 LF wafers  = w42-w48
# LAT LF wafers  = w39-w41
# LAT MF wafers  = w13-w24
# LAT HF wafers  = w00-w05

# Arbitrary mapping between wafer slots and array names
wafer_to_array = {
    "dummy00" : "Cv4",  # LAT
    "dummy01" : "Cv5",
    "w13" : "Mv6",   # LAT  1/12
    "w14" : "Mv7",   # LAT  2/12
    "dummy04" : "Mv9",
    "w15" : "Mv11",  # LAT  3/12
    "w16" : "Mv12",  # LAT  4/12
    "w25" : "Mv13",  # SAT1 1/7
    "w26" : "Mv14",  # SAT1 2/7
    # "Mv15",  # LAT (missing)
    "w17" : "Mv17",  # LAT  5/12
    "w27" : "Mv18",  # SAT1 3/7
    "w28" : "Mv19",  # SAT1 4/7
    "w29" : "Mv22",  # SAT1 5/7
    "w30" : "Mv23",  # SAT1 6/7
    "w31" : "Mv24",  # SAT1 7/7
    "w18" : "Mv25",  # LAT  6/12
    "w19" : "Mv26",  # LAT  7/12
    "w20" : "Mv27",  # LAT  8/12
    "w21" : "Mv28",  # LAT  9/12
    "w22" : "Mv29",  # LAT 10/12
    "w23" : "Mv32",  # LAT 11/12
    "w24" : "Mv33",  # LAT 12/12
    "dummy05" : "Sv5",
    "w06" : "Uv31", # Only one SAT HF wafer in DetMap
    "w07" : "Uv31", # Only one SAT HF wafer in DetMap
    "w08" : "Uv31", # Only one SAT HF wafer in DetMap
    "w09" : "Uv31", # Only one SAT HF wafer in DetMap
    "w10" : "Uv31", # Only one SAT HF wafer in DetMap
    "w11" : "Uv31", # Only one SAT HF wafer in DetMap
    "w12" : "Uv31", # Only one SAT HF wafer in DetMap
    "w00" : "Uv8",  # Only one LAT HF wafer in DetMap
    "w01" : "Uv8",  # Only one LAT HF wafer in DetMap
    "w02" : "Uv8",  # Only one LAT HF wafer in DetMap
    "w03" : "Uv8",  # Only one LAT HF wafer in DetMap
    "w04" : "Uv8",  # Only one LAT HF wafer in DetMap
    "w05" : "Uv8",  # Only one LAT HF wafer in DetMap
}



def pos_to_ind(pos, bp, pol, mapping, tol=1.0, verbose=False):
    """
        Match position to a detector index on the UFM

        pos: x, y coordinates (tuple) of first detector (mm) relative to UFM center
        bp: detector bandpass (int)
        pol: detector polarization (string), 'A' or 'B'
        mapping: pandas object from DetMap CSV of UFM characteristics
        tol (float):  Maximum allowed distance between provided and
            matched detector positions.
    """

    # Compute Pythagorean distances to all detectors in the mapping
    x, y = pos
    dist = np.sqrt(np.square(mapping["det_x"] - x) + np.square(mapping["det_y"] - y))

    # Send all out of bandpass and polarization distances to infinity
    dist[np.logical_or(mapping['bandpass'] != str(bp), mapping['pol'] != pol)] = np.inf

    # Match detector to position
    ind = dist.argmin()

    # Check to make sure the match is credible
    dist_min = dist.iloc[ind]
    if dist_min > tol:
        if verbose:
            msg = f"Failed to match ({x} mm, {y} mm) to a detector in "
            msg += f"mapping.  Minimum distance is {dist_min} mm"
            warnings.warn(msg)
        return None

    if verbose:
        print(f"Minimum distance is at {ind} : {dist_min} mm", flush=True)

    return ind


def pos_to_chi(focalplane, dets, alpha=9.64e-30, tol=1.0):
    """
        Calculate magnitude of crosstalk for all detector pairs given
        their physical positions and mapping

        focalplane (SOFocalplane):  Focalplane object
        dets (iterable):  detector names to consider
        alpha:  crosstalk prefactor (Hz^-2), from John Groh (via FastHenry),
            valid for nearest physical neighbors
        tol (float):  Maximum allowed distance between provided and
            matched detector positions.
    """
    if not detmap_available:
        raise RuntimeError("Cannot evaluate chi -- no DetMap available")

    # Get the bandpass, polarization and position for every detector
    bandpasses = []
    pols = []
    positions = []
    wafers = []
    for det in dets:
        wafers.append(focalplane[det]["wafer_slot"])
        bandpasses.append(int(focalplane[det]["band"][-3:]))
        pols.append(focalplane[det]["pol"])
        positions.append((
            focalplane[det]["wafer_x"].to_value(u.mm),
            focalplane[det]["wafer_y"].to_value(u.mm)
        ))

    # Load the appropriate mappings
    wafer_set = set(wafers)
    for wafer in wafer_set:
        if wafer not in wafer_to_array:
            msg = f"Could not map {wafer} to an array in DetMap. "
            msg += f"Mapped wafer slots are {sorted(wafer_to_array.keys())}"
            raise RuntimeError(msg)
        array = wafer_to_array[wafer]
        try:
            datafile = available_solutions.get_solution_file(array)
        except ValueError as e:
            msg = f"{array} does not appear to be a valid DetMap name:\n{e}\n "
            msg += f"Perhaps wafer_to_array is out of date?"
        mappings[wafer] = pd.read_csv(datafile)

    # chi for every detector pair
    chis = {}
    for wafer in wafer_set:
        # Collect detector and positions that match wafer
        m = mappings[wafer]
        if len(m) == 0:
            msg = f"No match in DetMap."
            raise RuntimeError(msg)
        detector_subset = []
        position_subset = []
        freq_subset = []
        is_north_subset = []
        mux_band_subset = []
        bond_pad_subset = []
        for d, w, b, p, pos in zip(dets, wafers, bandpasses, pols, positions):
            if w == wafer:
                detector_subset.append(d)
                position_subset.append(pos)
                ind = pos_to_ind(pos, b, p, m, tol=tol)
                if ind is not None:
                    # Get resonator frequency
                    freq_subset.append(m.iloc[ind]["freq_mhz"])
                    # Get other helpful variables
                    is_north_subset.append(m.iloc[ind]["is_north"])
                    mux_band_subset.append(m.iloc[ind]["mux_band"])
                    bond_pad_subset.append(m.iloc[ind]["bond_pad"])
                else:
                    # This position is not in the mapping.  For now, we
                    # keep the detector and disable crosstalk for it
                    freq_subset.append(0)
                    is_north_subset.append(None)
                    mux_band_subset.append(None)
                    bond_pad_subset.append(None)
        # Compute chi for all detector pairs in this subset
        ndet = len(detector_subset)
        for idet1, det1 in enumerate(detector_subset):
            x1, y1 = position_subset[idet1]
            freq1 = freq_subset[idet1]
            if freq1 == 0:
                # Detector was not found in the mapping and
                # there is no resonator frequency
                continue
            is_north1 = is_north_subset[idet1]
            mux_band1 = mux_band_subset[idet1]
            bond_pad1 = bond_pad_subset[idet1]
            for idet2 in range(idet1 + 1, ndet):
                freq2 = freq_subset[idet2]
                if freq2 == 0:
                    # Detector was not found in the mapping and
                    # there is no resonator frequency
                    continue
                det2 = detector_subset[idet2]
                is_north2 = is_north_subset[idet2]
                mux_band2 = mux_band_subset[idet2]
                bond_pad2 = bond_pad_subset[idet2]
                # Short-circuit chi-calculation if the detectors
                # cannot cross-talk
                if is_north1 != is_north2:
                    continue
                if mux_band1 != mux_band2:
                    continue
                if np.abs(bond_pad1 - bond_pad2) != 4:
                    continue
                x2, y2 = position_subset[idet2]
                # Translate frequencies to chi
                df = freq1 - freq2
                avg_f = (freq1 + freq2) / 2
                chi = alpha * np.power(avg_f, 4) * np.power(df, -2) * 1e12
                if chi > 1:
                    msg = f"Anomalously high chi at"
                    msg += f" {det1}@({x1}, {y1}), {det2}@({x2}, {y2})"
                    msg += f" : {chi}"
                    raise RuntimeError(msg)
                chis[(det1, det2)] = chi
                chis[(det2, det1)] = chi

    return chis



""" Test psd calculation
"""


import unittest
import numpy as np
from numpy.fft import rfftfreq, irfft

from sotodlib import core
from sotodlib.tod_ops import detrend_tod
from sotodlib.tod_ops.flags import get_turnaround_flags
from sotodlib.tod_ops.fft_ops import (
    calc_psd, calc_wn, fit_noise_model, noise_model)

from .test_azss import get_scan

TOL_BIAS = 0.005


class PSDTest(unittest.TestCase):
    def test_psd_fit(self):
        fs = 200.
        dets = core.LabelAxis('dets', [f'det{di:003}' for di in range(20)])
        nsamps = 200 * 3600

        aman = core.AxisManager(dets)
        ndets = aman.dets.count

        white_noise_amp_input = 50 + np.random.randn(ndets)  # W/sqrt{Hz}
        fknee_input = 1 + 0.1 * np.random.randn(ndets)
        alpha_input = 3 + 0.2 * np.random.randn(ndets)

        freqs = rfftfreq(nsamps, d=1/fs)
        params = [white_noise_amp_input[:, np.newaxis],
                  fknee_input[:, np.newaxis],
                  alpha_input[:, np.newaxis]]
        pxx_input = noise_model(freqs, params)

        pxx_input[:, 0] = 0

        T = nsamps/fs
        ft_amps = np.sqrt(pxx_input * T * fs**2 / 2)

        ft_phases = np.random.uniform(0, 2 * np.pi, size=ft_amps.shape)
        ft_coefs = ft_amps * np.exp(1.0j * ft_phases)
        realized_noise = irfft(ft_coefs)
        timestamps = 1700000000 + np.arange(0, realized_noise.shape[1])/fs
        aman.add_axis(core.OffsetAxis('samps', len(timestamps)))
        aman.wrap('timestamps', timestamps, [(0, 'samps')])
        aman.wrap('signal', realized_noise, [(0, 'dets'), (1, 'samps')])

        detrend_tod(aman)
        freqs_output, Pxx_output = calc_psd(aman, nperseg=200*100)
        fit_result = fit_noise_model(aman, wn_est=50, fknee_est=1.0,
                                     alpha_est=3.3, lowf=0.05,
                                     f_max=5, binning=True,
                                     psdargs={'nperseg': 200*1000})
        wnl_fit = fit_result.fit[:, 0]
        fk_fit = fit_result.fit[:, 1]
        alpha_fit = fit_result.fit[:, 2]

        self.assertTrue(np.abs(np.median(white_noise_amp_input - wnl_fit)) < 1)
        self.assertTrue(np.abs(np.median(fknee_input - fk_fit)) < 0.1)
        self.assertTrue(np.abs(np.median(alpha_input - alpha_fit)) < 0.1)

    def test_wn_debias(self):
        # prep
        timestamps, az = get_scan(
            n_scans=20, scan_accel=0.25, scanrate=0.5, az0=0, az1=40)

        nsamps = len(timestamps)
        ndets = 100
        np.random.seed(0)
        signal = np.random.normal(0, 1, size=(ndets, nsamps))

        dets = [f"det{i}" for i in range(ndets)]
        aman = core.AxisManager(
                core.LabelAxis("dets", dets),
                core.IndexAxis("samps", nsamps)
        )
        aman.wrap("timestamps", timestamps, [(0, "samps")])
        aman.wrap("signal", signal, [(0, "dets"), (1, "samps")])
        boresight = core.AxisManager(aman.samps)
        boresight.wrap("az", az, [(0, "samps")])
        aman.wrap('boresight', boresight)
        aman.wrap('flags', core.AxisManager(aman.dets, aman.samps))
        get_turnaround_flags(aman)

        # test default arguments, this is biased
        calc_psd(aman, merge=True, nperseg=2**18)
        wn = calc_wn(aman)
        ratio = np.average(wn) / np.sqrt(np.average(aman.Pxx))
        self.assertTrue(abs(ratio - 1) > TOL_BIAS)
        # test debias, full_output=True, noverlap=0
        freqs, Pxx, nseg = calc_psd(aman, merge=False, full_output=True,
                                    noverlap=0, nperseg=2**18)
        wn = calc_wn(aman, Pxx, freqs, nseg)
        ratio = np.average(wn) / np.sqrt(np.average(Pxx))
        self.assertAlmostEqual(ratio, 1, delta=TOL_BIAS)
        # test quarter nperseg
        freqs, Pxx, nseg = calc_psd(aman, merge=False, full_output=True,
                                    noverlap=0, nperseg=2**16)
        wn = calc_wn(aman, Pxx, freqs, nseg)
        ratio = np.average(wn) / np.sqrt(np.average(Pxx))
        self.assertAlmostEqual(ratio, 1, delta=TOL_BIAS)
        # test defulat nperseg
        freqs, Pxx, nseg = calc_psd(aman, merge=False, full_output=True,
                                    noverlap=0)
        wn = calc_wn(aman, Pxx, freqs, nseg)
        ratio = np.average(wn) / np.sqrt(np.average(Pxx))
        self.assertAlmostEqual(ratio, 1, delta=TOL_BIAS)
        # test subscan
        freqs, Pxx, nseg = calc_psd(aman, merge=False, full_output=True,
                                    noverlap=0, subscan=True)
        wn = calc_wn(aman, Pxx, freqs, nseg)
        ratio = np.average(wn) / np.sqrt(np.average(Pxx))
        self.assertAlmostEqual(ratio, 1, delta=TOL_BIAS)



def get_parser(parser=None):
    # a config file to pass all parameters is pending
    import argparse
    if parser is None:
        parser = argparse.ArgumentParser()
    parser.add_argument("query")
    parser.add_argument("area")
    parser.add_argument("odir")
    parser.add_argument("prefix", nargs="?")
    parser.add_argument(      "--comps",   type=str, default="TQU",help="List of components to solve for. T, QU or TQU, but only TQU is consistent with the actual data")
    parser.add_argument("-W", "--wafers",  type=str, default=None, help="Detector wafer subsets to map with. ,-sep")
    parser.add_argument("-B", "--bands",   type=str, default=None, help="Bandpasses to map. ,-sep")
    parser.add_argument("-C", "--context", type=str, default="/mnt/so1/shared/todsims/pipe-s0001/v4/context.yaml")
    parser.add_argument(      "--tods",    type=str, default=None, help="Arbitrary slice to apply to the list of tods to analyse")
    parser.add_argument("-n", "--ntod",    type=int, default=None, help="Keep at most this many tods")
    parser.add_argument("-N", "--nmat",    type=str, default="corr", help="Noise model to use. corr or uncorr")
    parser.add_argument(      "--max-dets",type=int, default=None,   help="Keep at most this many detectors")
    parser.add_argument("-S", "--site",    type=str, default="so_lat")
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("-q", "--quiet",   action="count", default=0)
    parser.add_argument("-@", "--center-at", type=str, default=None)
    parser.add_argument("-w", "--window",  type=float, default=0.0)
    parser.add_argument("-i", "--inject",  type=str,   default=None, help="Path to map to inject. Equatorial coordinates")
    parser.add_argument(      "--nocal",   action="store_true", help="Disable calibration. Useful for sims")
    parser.add_argument(      "--nmat-dir",  type=str, default="{odir}/nmats")
    parser.add_argument(      "--nmat-mode", type=str, default="build", help="How to build the noise matrix. 'build': Always build from tod. 'cache': Use if available in nmat-dir, otherwise build and save. 'load': Load from nmat-dir, error if missing. 'save': Build from tod and save.")
    parser.add_argument("-d", "--downsample", type=str, default="1",  help="Downsample TOD by this factor. ,-sep")
    parser.add_argument(      "--maxiter",    type=str, default="500",help="Max number of CG steps per pass. ,-sep")
    parser.add_argument(      "--interpol",   type=str, default="nearest", help="Pmat interpol per pass. ,-sep")
    parser.add_argument("-T", "--tiled"  ,   type=int, default=1, help="0: untiled maps. Nonzero: tiled maps")
    parser.add_argument(      "--srcsamp",   type=str, default=None, help="path to mask file where True regions indicate where bright object mitigation should be applied. Mask is in equatorial coordinates. Not tiled, so should be low-res to not waste memory.")
    parser.add_argument(      "--unit",      type=str, default="uK", help="Unit of the maps")
    return parser

def main(**args):
    import numpy as np, sys, time, warnings, os, so3g
    from sotodlib.core import Context, AxisManager, IndexAxis
    from sotodlib.io import metadata   # PerDetectorHdf5 work-around
    from sotodlib import tod_ops, mapmaking, core
    from sotodlib.tod_ops import filters
    from sotodlib.mapmaking import log
    from pixell import enmap, utils, fft, bunch, wcsutils, mpi, bench
    import yaml

    #try: import moby2.analysis.socompat
    #except ImportError: warnings.warn("Can't import moby2.analysis.socompat. ACT data input probably won't work")
    class DataMissing(Exception): pass

    warnings.simplefilter('ignore')
    args    = bunch.Bunch(**args)
    SITE    = args.site.lower()
    verbose = args.verbose - args.quiet
    comm    = mpi.COMM_WORLD
    shape, wcs = enmap.read_map_geometry(args.area)

    # Reconstruct that wcs in case default fields have changed; otherwise
    # we risk adding information in MPI due to reconstruction, and that
    # can cause is_compatible failures.
    wcs = wcsutils.WCS(wcs.to_header())
    # Set shape to None to allow map to fit these TODs exactly.
    #shape = None

    comps = args.comps
    ncomp = len(comps)
    dtype_tod = np.float32
    dtype_map = np.float64
    nmat_dir  = args.nmat_dir.format(odir=args.odir)
    prefix= args.odir + "/"
    if args.prefix: prefix += args.prefix + "_"
    utils.mkdir(args.odir)
    L = mapmaking.init(level=mapmaking.DEBUG, rank=comm.rank)

    recenter = None
    if args.center_at:
        recenter = mapmaking.parse_recentering(args.center_at)

    with bench.mark('context'):
        context = Context(args.context)

    wafers  = args.wafers.split(",") if args.wafers else None
    bands   = args.bands .split(",") if args.bands  else None
    sub_ids = mapmaking.get_subids(args.query, context=context)
    sub_ids = mapmaking.filter_subids(sub_ids, wafers=wafers, bands=bands)

    # restrict tod selection further. E.g. --tods [0], --tods[:1], --tods[::100], --tods[[0,1,5,10]], etc.
    if args.tods:
        sub_ids = eval("sub_ids" + args.tods)
    if args.ntod is not None:
        sub_ids = sub_ids[:args.ntod]

    if len(sub_ids) == 0:
        if comm.rank == 0:
            print("No tods found!")
        sys.exit(1)
    L.info("Found %d tods" % (len(sub_ids)))

    if args.inject:
        map_to_inject = enmap.read_map(args.inject).astype(dtype_map)

    if args.srcsamp:
        srcsamp_mask  = enmap.read_map(args.srcsamp)

    passes = mapmaking.setup_passes(downsample=args.downsample, maxiter=args.maxiter, interpol=args.interpol)
    for ipass, passinfo in enumerate(passes):
        L.info("Starting pass %d/%d maxit %d down %d interp %s" % (ipass+1, len(passes), passinfo.maxiter, passinfo.downsample, passinfo.interpol))
        pass_prefix = prefix + "pass%d_" % (ipass+1)
        # Multipass mapmaking
        # Will do multiple passes over the data, using the previous pass both as
        # the initial condition and to temporarily subtract when estimating the noise
        # model. Typically the different passes will differ by sample rate. This means
        # that e.g. SignalCut will be different for each pass, and the degrees of freedom
        # will need to be translated between them. Rough sketch:
        # 1. each pass defines new signals and new mapmaker
        # 2. loop through tods and preprare them as normal
        # 3. just before mapmaker.add_obs, translate the old solution and evaluate it to a tod for
        #    this observation. Pass this as an optional parameter to add_obs, since that's where the
        #    noise model is built. In add_obs, this is subtracted from the tod before the noise model is
        #    built, and then added back again.
        # 4. Separately, build up the transated degrees of freedom. How can we do this without
        #    hardcoding stuff for specific signals? Some signals have per-tod degrees of freedom,
        #    others don't. Each signal should probably have an x = signal.translate(signal_old, x_old) method.
        #    SignalCut already knows about the tod size and cut ranges. Then the mapmaker can also have
        #    an x = mapmaker.translate(mapmaker_old, x_old) method, which assumes a compatible mapmaker,
        #    and basically does
        #      xvals = []
        #      for signal, signal_old, xvals_old in zip(mapmaker.signals, mapmaker_old.signals, mapmaker_old.dof.unzip(x_old)):
        #         xvals.append(signal.translate(signal_old, xvals))
        #      return mapmaker.dof.zip(*xvals)
        #    Except we can't do this before we've added all the tods, because thew new mapmaker's dofs won't have
        #    been set up yet! Well, just do it afterwards, then. That's fine for #4, but it means we can't
        #    use this to implement #3.
        #
        # I need a way to translate and evaluate a signal for a single tod, before the whole signal has
        # been constructed, or this tod has even been added to it (since the adding happens after
        # the noise model is built). Probably best to just make a separate function for this, which doesn't
        # need to know about the new degrees of freedom, which haven't been finalized at this point. This
        # would end up doing some redundant operations, e.g. defining the map pointing matrix twice, but
        # I don't see a cleaner design than this.
        # This could have an interface like tod = signal.transeval(signal_old, id, obs, xval), or it could be
        # split into two parts signal.translate_single and signal.forward_single). But I don't think those
        # building blocks would be very reusable, and the full thing is more general.
        if   args.nmat == "uncorr": noise_model = mapmaking.NmatUncorr()
        elif args.nmat == "corr":   noise_model = mapmaking.NmatDetvecs(verbose=verbose>1, downweight=[1e-4, 0.25, 0.50], window=args.window)
        else: raise ValueError("Unrecognized noise model '%s'" % args.nmat)

        signal_cut = mapmaking.SignalCut(comm, dtype=dtype_tod)
        signal_map = mapmaking.SignalMap(shape, wcs, comm, comps=comps, dtype=dtype_map, recenter=recenter, tiled=args.tiled>0, interpol=args.interpol)
        signals    = [signal_cut, signal_map]
        if args.srcsamp:
            signal_srcsamp = mapmaking.SignalSrcsamp(comm, srcsamp_mask, dtype=dtype_tod)
            signals.append(signal_srcsamp)
        mapmaker   = mapmaking.MLMapmaker(signals, noise_model=noise_model, dtype=dtype_tod, verbose=verbose>0)

        nkept = 0
        # TODO: Fix the task distribution. The current one doesn't care which mpi
        # task gets which tods, which sabotages the pixel-saving effects of tiled maps!
        # To be able to distribute the tods sensibly, we need a rough estimate of where
        # on the sky each tod is. We should be able to get this using the central
        # ctime, az and el for each tod.
        for ind in range(comm.rank, len(sub_ids), comm.size):
            # Detsets correspond to separate files, so treat them as separate TODs.
            sub_id = sub_ids[ind]
            obs_id, wafer, band = sub_id.split(":")
            name = sub_id.replace(":", "_")
            L.debug("Processing %s" % sub_id)
            try:
                meta = context.get_meta(sub_id)
                # Optionally restrict to maximum number of detectors. This is mainly
                # useful for doing fast debug runs. Before doing this we make sure to
                # sort the detctor list so we chop off a deterministic subset of detectors.
                meta.restrict("dets", np.sort(meta.dets.vals))
                my_dets = meta['dets'].vals
                if args.max_dets is not None:
                    meta.restrict('dets', meta['dets'].vals[:args.max_dets])
                if len(my_dets) == 0: raise DataMissing("no dets left")
                # Actually read the data
                with bench.mark("read_obs %s" % sub_id):
                    obs = context.get_obs(sub_id, meta=meta)

                # Fix boresight
                mapmaking.fix_boresight_glitches(obs)
                # Get our sample rate. Would have been nice to have this available in the axisman
                srate = (obs.samps.count-1)/(obs.timestamps[-1]-obs.timestamps[0])

                # Add site and weather, since they're not in obs yet
                obs.wrap("weather", np.full(1, "typical"))
                obs.wrap("site",    np.full(1, SITE))

                # Prepare our data. FFT-truncate for faster fft ops
                obs.restrict("samps", [0, fft.fft_len(obs.samps.count)])

                # Desolope to make it periodic. This should be done *before*
                # dropping to single precision, to avoid unnecessary loss of precision due
                # to potential high offses in the raw tod.
                utils.deslope(obs.signal, w=5, inplace=True)
                obs.signal = obs.signal.astype(dtype_tod)

                if "flags" not in obs:
                    obs.wrap("flags", core.AxisManager(obs.dets, obs.samps))

                if "glitch_flags" not in obs.flags:
                    if "glitch_flags" in obs:
                        obs.flags.wrap("glitch_flags", obs.glitch_flags, axis_map=((0,obs.dets),(1,obs.samps)))
                    else:
                        obs.flags.wrap_new('glitch_flags', shape=('dets', 'samps'), cls=so3g.proj.RangesMatrix.zeros)

                # Optionally skip all the calibration. Useful for sims.
                if not args.nocal:
                    # Disqualify overly cut detectors
                    good_dets = mapmaking.find_usable_detectors(obs)
                    obs.restrict("dets", good_dets)
                    if obs.dets.count == 0:
                        L.debug("Skipped %s (all dets cut)" % (sub_id))
                        continue
                    # Gapfill glitches. This function name isn't the clearest
                    tod_ops.get_gap_fill(obs, flags=obs.flags.glitch_flags, swap=True)
                    # Gain calibration
                    gain  = 1
                    for gtype in ["relcal","abscal"]:
                        gain *= obs[gtype][:,None]
                    obs.signal *= gain
                    # Fourier-space calibration
                    fsig  = fft.rfft(obs.signal)
                    freq  = fft.rfftfreq(obs.samps.count, 1/srate)
                    # iir filter
                    iir_filter  = filters.iir_filter()(freq, obs)
                    fsig       /= iir_filter
                    gain       /= iir_filter[0].real # keep track of total gain for our record
                    fsig       /= filters.timeconst_filter(None)(freq, obs)
                    fft.irfft(fsig, obs.signal, normalize=True)
                    del fsig

                    # Apply pointing correction.
                    #obs.focal_plane.xi    += obs.boresight_offset.xi
                    #obs.focal_plane.eta   += obs.boresight_offset.eta
                    #obs.focal_plane.gamma += obs.boresight_offset.gamma
                    obs.focal_plane.xi    += obs.boresight_offset.dx
                    obs.focal_plane.eta   += obs.boresight_offset.dy
                    obs.focal_plane.gamma += obs.boresight_offset.gamma

                # Injecting at this point makes us insensitive to any bias introduced
                # in the earlier steps (mainly from gapfilling). The alternative is
                # to inject it earlier, and then anti-calibrate it.
                # Might want to make the interpol used here separate from the one in
                # the main mapmaking.
                if args.inject:
                    mapmaking.inject_map(obs, map_to_inject, recenter=recenter, interpol=args.interpol)
                utils.deslope(obs.signal, w=5, inplace=True)

                if passinfo.downsample != 1:
                    obs = mapmaking.downsample_obs(obs, passinfo.downsample)

                # Maybe load precomputed noise model.
                # FIXME: How to handle multipass here?
                nmat_file = nmat_dir + "/nmat_%s.hdf" % name
                if args.nmat_mode == "load" or args.nmat_mode == "cache" and os.path.isfile(nmat_file):
                    L.debug("Reading noise model %s" % nmat_file)
                    nmat = mapmaking.read_nmat(nmat_file)
                else: nmat = None

                # And add it to the mapmaker
                with bench.mark("add_obs %s" % sub_id):
                    if ipass > 0:
                        # Evaluate the final model of the previous pass' mapmaker
                        # for this observation.
                        signal_estimate = eval_prev.evaluate(mapmaker_prev.data[len(mapmaker.data)])
                        # Resample this to the current downsampling level
                        signal_estimate = mapmaking.resample.resample_fft_simple(signal_estimate, obs.samps.count)
                    else: signal_estimate = None
                    mapmaker.add_obs(sub_id, obs, noise_model=nmat, signal_estimate=signal_estimate)
                    del signal_estimate
                del obs
                nkept += 1

                # Maybe save the noise model we built (only if we actually built one rather than
                # reading one in)
                if args.nmat_mode in ["save", "cache"] and nmat is None:
                    print("Writing noise model %s" % nmat_file)
                    utils.mkdir(nmat_dir)
                    mapmaking.write_nmat(nmat_file, mapmaker.data[-1].nmat)
            except (DataMissing,IndexError,ValueError) as e:
                L.debug("Skipped %s (%s)" % (sub_id, str(e)))
                continue

        nkept = comm.allreduce(nkept)
        if nkept == 0:
            if comm.rank == 0:
                L.info("All tods failed. Giving up")
            sys.exit(1)

        L.info("Done building")

        with bench.mark("prepare"):
            mapmaker.prepare()
        L.info("Done preparing")

        signal_map.write(prefix, "rhs", signal_map.rhs, unit=args.unit+'^-1')
        signal_map.write(prefix, "div", signal_map.div, unit=args.unit+'^-2')
        signal_map.write(prefix, "bin", enmap.map_mul(signal_map.idiv, signal_map.rhs), unit=args.unit)
        L.info("Wrote rhs, div, bin")

        # Set up initial condition
        x0 = None if ipass == 0 else mapmaker.translate(mapmaker_prev, eval_prev.x_zip)

        t1 = time.time()
        for step in mapmaker.solve(maxiter=passinfo.maxiter, x0=x0):
            t2 = time.time()
            dump = step.i % 10 == 0
            L.info("CG step %4d %15.7e %8.3f %s" % (step.i, step.err, t2-t1, "" if not dump else "(write)"))
            if dump:
                for signal, val in zip(signals, step.x):
                    if signal.output:
                        signal.write(pass_prefix, "map%04d" % step.i, val)
            t1 = time.time()

        L.info("Done")
        for signal, val in zip(signals, step.x):
            if signal.output:
                signal.write(pass_prefix, "map", val, unit=args['unit'])
        comm.Barrier()

        mapmaker_prev = mapmaker
        eval_prev     = mapmaker.evaluator(step.x_zip)

if __name__ == '__main__':
    from sotodlib.site_pipeline import util
    util.main_launcher(main, get_parser)



"""
This module is used to compute detector calibration parameters from sodetlib
data products.

The naive computation is described in the `sodetlib documentation. <https://sodetlib.readthedocs.io/en/latest/operations/bias_steps.html#in-transition>`_

Details about the RP and loopgain correction `can be found on our confluence. <https://simonsobs.atlassian.net/wiki/spaces/~5570586d07625a6be74c8780e4b96f6156f5e6/blog/2024/02/02/286228683/Nonlinear+TES+model+using+RP+curve>`_
"""

import traceback
import os
import yaml
from dataclasses import dataclass, astuple, fields
import numpy as np
from tqdm.auto import tqdm
import logging
from typing import Optional, Union, Dict, List, Any, Tuple, Literal, Callable
from queue import Queue
import argparse

from so3g.proj import RangesMatrix
from sotodlib import core
from sotodlib.io.metadata import write_dataset, ResultSet
from sotodlib.io.load_book import get_cal_obsids
from sotodlib.utils.procs_pool import get_exec_env
from sotodlib.hwp import get_hwpss, subtract_hwpss
import sotodlib.site_pipeline.util as sp_util
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed, Future
import sodetlib.tes_param_correction as tpc
from sodetlib.operations.iv import IVAnalysis
from sodetlib.operations.bias_steps import BiasStepAnalysis


# stolen  from pysmurf, max bias volt / num_bits
DEFAULT_RTM_BIT_TO_VOLT = 10 / 2**19
DEFAULT_pA_per_phi0 = 9e6
TES_BIAS_COUNT = 12  # per detset / primary file group

# For converting bias group to bandpass.
BGS = {'lb': [0, 1, 4, 5, 8, 9], 'hb': [2, 3, 6, 7, 10, 11]}
BAND_STR = {'mf': {'lb': 'f090', 'hb': 'f150'},
            'uhf': {'lb': 'f220', 'hb': 'f280'},
            'lf': {'lb': 'f030', 'hb': 'f040'}}

logger = logging.getLogger("det_cal")
if not logger.hasHandlers():
    sp_util.init_logger("det_cal")


def get_data_root(ctx: core.Context, obs_id:str) -> str:
    "Get root data directory based on context file"
    c = ctx.obsfiledb.conn.execute("select name from files limit 1")
    res = [r[0] for r in c][0]
    # split out <data_root>/obs/<timecode>/<obsid>/fname
    for _ in range(5):
        res = os.path.dirname(res)
    # add telescope str to path
    tel_str = obs_id.split(('_'))[2]
    res = os.path.join(res, tel_str)
    return res


@dataclass
class DetCalCfg:
    """
    Class for configuring the behavior of the det-cal update script.

    Args
    -------------
    root_dir: str
        Path to the root of the results directory.
    context_path: str
        Path to the context file to use.
    raise_exceptions: bool
        If Exceptions should be raised in the get_cal_resset function.
        Defaults to False.
    apply_cal_correction: bool
        If True, apply the RP calibration correction, and use corrected results
        for Rtes, Si, Pj, and loopgain when successful. Defaults to True.
    index_path: str
        Path to the index file to use for the det_cal database. Defaults to
        "det_cal.sqlite".
    h5_path: str
        Path to the HDF5 file to use for the det_cal database. Default to
        "det_cal.h5".
    h5_unix_digits: int
        Number of digits of unixtime to be added to h5_path. For example if
        h5_unix_digits = 4, h5_path will be modified to "det_cal_1700.h5".
        Defaults to 0.
    cache_failed_obsids: bool
        If True, will cache failed obs-ids to avoid re-running them. Defaults to
        True.
    failed_file_cache: str
        Path to the yaml file that will store failed obsids. Defaults to
        "failed_obsids.yaml".
    show_pb: bool
        If True, show progress bar in the run_update function. Defaults to True.
    param_correction_config: dict
        Configuration for the TES param correction. If None, default values are used.
    run_method: str
        Must be "site" or "nersc". If "site", this function will not parallelize SQLite access, and will only parallelize the TES parameter correction. If "nersc", this will parallelize both SQLite access and the TES param correction, using ``nprocs_obs_info`` and ``nprocs_result_set`` processes respectively.
    nprocs_obs_info: int
        Number of processes to use to acquire observation info from the file system.
        Defaults to 1.
    nprocs_result_set: int
        Number of parallel processes that should to compute the TES parameters,
        and to run the TES parameter correction.
    num_obs: Optional[int]
        Max number of observations to process per run_update call. If not set,
        will run on all available observations.
    log_level: str
        Logging level for the logger.
    multiprocess_start_method: str
        Method to use to start child processes. Can be "spawn" or "fork".
    """

    def __init__(
        self,
        root_dir: str,
        context_path: str,
        *,
        raise_exceptions: bool = False,
        apply_cal_correction: bool = True,
        hwpss_subtraction: bool = False,
        metadata_list: Union[str, List[str]] = 'all',
        index_path: str = "det_cal.sqlite",
        h5_path: str = "det_cal.h5",
        h5_unix_digits: int = 0,
        cache_failed_obsids: bool = True,
        failed_cache_file: str = "failed_obsids.yaml",
        show_pb: bool = True,
        param_correction_config: Union[Dict[str, Any], None, tpc.AnalysisCfg] = None,
        run_method: str = "site",
        nprocs_obs_info: int = 1,
        nprocs_result_set: int = 10,
        num_obs: Optional[int] = None,
        log_level: str = "DEBUG",
        multiprocess_start_method: Literal["spawn", "fork"] = "spawn"
    ) -> None:
        self.root_dir = root_dir
        self.context_path = os.path.expandvars(context_path)
        self.metadata_list = metadata_list
        self.raise_exceptions = raise_exceptions
        self.apply_cal_correction = apply_cal_correction
        self.hwpss_subtraction = hwpss_subtraction
        self.cache_failed_obsids = cache_failed_obsids
        self.show_pb = show_pb
        self.run_method = run_method

        if self.run_method not in ["site", "nersc"]:
            raise ValueError("run_method must be in: ['site', 'nersc']")

        self.nprocs_obs_info = nprocs_obs_info
        self.nprocs_result_set = nprocs_result_set
        self.num_obs = num_obs
        self.log_level = log_level
        self.multiprocess_start_method = multiprocess_start_method

        self.root_dir = os.path.expandvars(self.root_dir)
        if not os.path.exists(self.root_dir):
            raise ValueError(f"Root dir does not exist: {self.root_dir}")

        def parse_path(path: str) -> str:
            "Expand vars and make path absolute"
            p = os.path.expandvars(path)
            if not os.path.isabs(p):
                p = os.path.join(self.root_dir, p)
            return p

        self.index_path = parse_path(index_path)
        self.h5_path = parse_path(h5_path)
        self.h5_unix_digits = h5_unix_digits
        self.failed_cache_file = parse_path(failed_cache_file)

        kw = {"show_pb": False, "default_nprocs": self.nprocs_result_set}
        if param_correction_config is None:
            self.param_correction_config = tpc.AnalysisCfg(**kw)  # type: ignore
        elif isinstance(param_correction_config, dict):
            kw.update(param_correction_config)
            self.param_correction_config = tpc.AnalysisCfg(**kw)  # type: ignore
        else:
            self.param_correction_config = param_correction_config

        self.setup_files()

    @classmethod
    def from_yaml(cls, path) -> "DetCalCfg":
        with open(path, "r") as f:
            d = yaml.safe_load(f)
            return cls(**d)

    def setup_files(self) -> None:
        """Create directories and databases if they don't exist"""
        if not os.path.exists(self.failed_cache_file):
            # If file doesn't exist yet, just create an empty one
            with open(self.failed_cache_file, "w") as f:
                yaml.dump({}, f)

        if not os.path.exists(self.index_path):
            scheme = core.metadata.ManifestScheme()
            scheme.add_exact_match("obs:obs_id")
            scheme.add_data_field("dataset")
            db = core.metadata.ManifestDb(scheme=scheme)
            db.to_file(self.index_path)


@dataclass
class CalInfo:
    """
    Class that contains detector calibration information that will go into
    the caldb.

    Attributes
    ----------
    readout_id: str
        Readout ID of the detector.
    r_tes: float
        Detector resistance [ohms], determined through bias steps while the
        detector is biased.
    r_frac: float
        Fractional resistance of TES, given by r_tes / r_n.
    p_bias: float
        Bias power on the TES [W] computed using bias steps at the bias point.
    s_i: float
        Current responsivity of the TES [1/V] computed using bias steps at the
        bias point.
    phase_to_pW: float
        Phase to power conversion factor [pW/rad] computed using s_i,
        pA_per_phi0, and detector polarity.
    v_bias: float
        Commanded bias voltage [V] on the bias line of the detector for the observation.
    tau_eff: float
        Effective thermal time constant [sec] of the detector, measured from bias steps.
    loopgain: float
        Loopgain of the detector.
    tes_param_correction_success: bool
        True if TES parameter corrections were successfully applied.
    bg: int
        Bias group of the detector. Taken from IV curve data, which contains
        bgmap data taken immediately prior to IV. This will be -1 if the
        detector is unassigned.
    polarity: int
        Polarity of the detector response for a positive change in bias current
        while the detector is superconducting. This is needed to correct for
        detectors that have reversed response.
    r_n: float
        Normal resistance of the TES [Ohms] calculated from IV curve data.
    p_sat: float
        "saturation power" of the TES [W] calculated from IV curve data.
        This is defined  as the electrical bias power at which the TES
        resistance is 90% of the normal resistance.
    naive_r_tes: float
        Detector resistance [ohms]. This is based on the naive bias step
        estimation without any additional corrections.
    naive_r_frac: float
        Fractional resistance of TES, given by r_tes / r_n. This is based on the
        naive bias step estimation without any additional corrections.
    naive_p_bias: float
        Bias power on the TES [W] computed using bias steps at the bias point.
        This is based on the naive bias step estimation without any additional
        corrections.
    naive_s_i: float
        Current responsivity of the TES [1/V] computed using bias steps at the
        bias point. This is based on the naive bias step estimation without
        using any additional corrections.
    bandpass: str
        Detector bandpass, computed from bias group information.
    """

    readout_id: str = ""
    r_tes: float = np.nan
    r_frac: float = np.nan
    p_bias: float = np.nan
    s_i: float = np.nan
    phase_to_pW: float = np.nan
    v_bias: float = np.nan
    tau_eff: float = np.nan
    loopgain: float = np.nan
    tes_param_correction_success: bool = False
    bg: int = -1
    polarity: int = 1
    r_n: float = np.nan
    p_sat: float = np.nan

    naive_r_tes: float = np.nan
    naive_r_frac: float = np.nan
    naive_p_bias: float = np.nan
    naive_s_i: float = np.nan
    bandpass: str = "NC"

    @classmethod
    def dtype(cls) -> List[Tuple[str, Any]]:
        """Returns ResultSet dtype for an item based on this class"""
        dtype = []
        for field in fields(cls):
            if field.name == "readout_id":
                dt: Tuple[str, Any] = ("dets:readout_id", "<U40")
            elif field.name == 'bandpass':
                # Our bandpass str is max 4 characters
                dt: Tuple[str, Any] = ("bandpass", "<U4")
            else:
                dt = (field.name, field.type)
            dtype.append(dt)
        return dtype


@dataclass
class ObsInfo:
    """
    Class containing observation gathered from obsdbs and the file system
    required to compute calibration results.

    Attributes
    ------------
    obs_id: str
        Obs id.
    iv_obsids: dict
        Dict mapping detset to iv obs-id.
    bs_obsids: dict
        Dict mapping detset to bias step obs-id.
    iva_files: dict
        Dict mapping detset to IV analysis file path.
    bsa_files: dict
        Dict mapping detset to bias step analysis file path.
    det_info: dict
        Dict which contains det readout id, band, channel, stream id, detset.
    bias_lines: list
        List which stores bias lines
    biases: ndarray
        Array of detector biases for each bias line
    tube_flavor: str
        Type of optics tube (MF, LF, or UHF)
    """

    obs_id: str

    iv_obsids: Dict[str, str]
    bs_obsids: Dict[str, str]

    iva_files: Dict[str, str]
    bsa_files: Dict[str, str]

    det_info: Dict[str, List[str]]
    bias_lines: List[str]
    biases: np.ndarray
    tube_flavor: str


@dataclass
class ObsInfoResult:
    obs_id: str
    success: bool = False
    traceback: str = ""
    obs_info: Optional[ObsInfo] = None


def get_obs_info(cfg: DetCalCfg, obs_id: str) -> ObsInfoResult:
    res = ObsInfoResult(obs_id)

    try:
        ctx = core.Context(cfg.context_path, metadata_list=cfg.metadata_list)
        am = ctx.get_obs(
            obs_id,
            samples=(0, 1),
            ignore_missing=True,
            no_signal=True,
            no_headers=False,
            on_missing={"det_cal": "skip"},
        )

        if "smurf" not in am.det_info:
            raise ValueError(f"Missing smurf info for {obs_id}")

        logger.debug(f"Getting cal obsids ({obs_id})")

        iv_obsids = get_cal_obsids(ctx, obs_id, "iv")

        # Load in IVs
        logger.debug(f"Loading Bias step and IV data ({obs_id})")
        rtm_bit_to_volt = None
        pA_per_phi0 = None

        # Automatically determine paths based on data root instead of obsfiledb
        # because obsfiledb queries are slow on nersc.

        iva_files = {}
        bsa_files = {}
        for dset, oid in iv_obsids.items():
            if oid is not None:
                timecode = oid.split("_")[1][:5]
                zsmurf_dir = os.path.join(
                    get_data_root(ctx, obs_id), "oper", timecode, oid, f"Z_smurf"
                )
                for f in os.listdir(zsmurf_dir):
                    if "iv" in f:
                        iva_files[dset] = os.path.join(zsmurf_dir, f)
                        break
                else:
                    raise ValueError(f"IV data not found for in cal obs {oid}")
            else:
                logger.debug("missing IV data for %s", dset)

        if len(iva_files) == 0:
            raise ValueError(f"No IV data found for {obs_id}")

        # Load in bias steps
        bias_step_obsids = get_cal_obsids(ctx, obs_id, "bias_steps")
        for dset, oid in bias_step_obsids.items():
            if oid is not None:
                timecode = oid.split("_")[1][:5]
                zsmurf_dir = os.path.join(
                    get_data_root(ctx, obs_id), "oper", timecode, oid, f"Z_smurf"
                )
                for f in os.listdir(zsmurf_dir):
                    if "bias_step" in f:
                        bs_file = os.path.join(zsmurf_dir, f)
                        bsa_files[dset] = bs_file
                        break
                else:
                    raise ValueError(f"Bias step data not found for in cal obs {oid}")
            else:
                logger.debug("missing bias step data for %s", dset)

        if rtm_bit_to_volt is None:
            rtm_bit_to_volt = DEFAULT_RTM_BIT_TO_VOLT
        if pA_per_phi0 is None:
            pA_per_phi0 = DEFAULT_pA_per_phi0

        # Pass through specific fields from our axismanager so that the result
        # is Picklable.
        res.obs_info = ObsInfo(
            obs_id=obs_id,
            det_info ={"stream_id": am.det_info.stream_id,
                       "detset": am.det_info.detset,
                       "readout_id": am.det_info.readout_id,
                       "band": am.det_info.smurf.band,
                       "channel": am.det_info.smurf.channel,},
            bias_lines=am.bias_lines.vals,
            biases=am.biases,
            tube_flavor=am.obs_info.tube_flavor,

            iv_obsids=iv_obsids,
            bs_obsids=bias_step_obsids,
            iva_files=iva_files,
            bsa_files=bsa_files,
        )
        res.success = True
    except:
        res.traceback = traceback.format_exc()
        if cfg.raise_exceptions:
            raise
    return res


@dataclass
class CalRessetResult:
    """
    Results object for the get_cal_resset function.
    """

    obs_info: ObsInfo
    success: bool = False
    traceback: Optional[str] = None
    fail_msg: Optional[str] = None

    correction_results: Optional[Dict[str, List[tpc.CorrectionResults]]] = None
    result_set: Optional[np.ndarray] = None


def biases_flags(bsa, buffer=200):
    """
    Make flags that mask bias steps

    Args
        bsa: sodetlib BiasStepAnalysis object
        buffer: Number of samples to buffer flags
    Returns
        RangesMatrix
    """
    mask = np.zeros((bsa.am.dets.count, bsa.am.samps.count),
                    dtype=bool)
    for i, bg in enumerate(bsa.bgmap):
        if bg == -1:
            continue
        mask[i][bsa.edge_idxs[bg][0]:bsa.edge_idxs[bg][-1]] = 1
    flags = RangesMatrix.from_mask(mask).buffer(buffer)
    return flags


def load_and_reanalyze_bs(bsa, ctx, obs_id):
    """
    Load raw data of biassteps and reanalyze it with hwpss subtraction

    Args
        bsa: sodetlib BiasStepAnalysis object
        ctx: Context object
        obs_id: observation id of bias steps
    """
    am = ctx.get_obs(obs_id)
    am.wrap('hwp_angle', am.hwp_solution.hwp_angle,
            [(0, 'samps')])
    if np.all(am.hwp_angle == 0):
        return
    bsa.am = am
    bsa._find_bias_edges()
    flags = biases_flags(bsa)
    get_hwpss(am, flags=flags, merge_stats=True)
    subtract_hwpss(am, subtract_name='signal')
    bsa._get_step_response()
    bsa._compute_dc_params()
    bsa._fit_tau_effs()
    del bsa.am


def get_cal_resset(cfg: DetCalCfg, obs_info: ObsInfo,
                   executor=None, as_completed_callable=None) -> CalRessetResult:
    """
    Returns calibration ResultSet for a given ObsId. This pulls IV and bias step
    data for each detset in the observation, and uses that to compute CalInfo
    for each detector in the observation.

    Args
    ------
    cfg: DetCalCfg
        DetCal configuration object.
    obs_info: ObsInfo
        ObsInfo object.
    pool: Optional[multiprocessing.Pool]
        If specified, will run TES param correction in parallel using processes
        from this pool.
    """

    obs_id = obs_info.obs_id
    res = CalRessetResult(obs_info)
    logger.debug("Computing Result set for %s", obs_info.obs_id)

    # Need to reset logger here because this may be created new for spawned process
    logger.setLevel(getattr(logging, cfg.log_level.upper()))
    for ch in logger.handlers:
        ch.setLevel(getattr(logging, cfg.log_level.upper()))

    try:

        ivas = {
            dset: IVAnalysis.load(iva_file)
            for dset, iva_file in obs_info.iva_files.items()
        }
        bsas = {
            dset: BiasStepAnalysis.load(bsa_file)
            for dset, bsa_file in obs_info.bsa_files.items()
        }

        if cfg.apply_cal_correction:
            for iva in ivas.values():
                # Run R_L correction if analysis version is old...
                if getattr(iva, "analysis_version", 0) == 0:
                    # This will edit IVA dicts in place
                    logger.debug("Recomputing IV analysis for %s", obs_id)
                    tpc.recompute_ivpars(iva, cfg.param_correction_config)

        if cfg.hwpss_subtraction:
            # Reanalyze biasstep with hwpss subtraction
            ctx = core.Context(cfg.context_path, metadata_list=cfg.metadata_list)
            bias_step_obsids = get_cal_obsids(ctx, obs_id, "bias_steps")

            for dset, bsa in bsas.items():
                oid = bias_step_obsids[dset]
                load_and_reanalyze_bs(bsa, ctx, oid)

        iva = list(ivas.values())[0]
        rtm_bit_to_volt = iva.meta["rtm_bit_to_volt"]
        pA_per_phi0 = iva.meta["pA_per_phi0"]
        cals = [CalInfo(rid) for rid in obs_info.det_info["readout_id"]]
        if len(cals) == 0:
            raise ValueError(f"No detectors found for {obs_id}")

        # Add IV info
        for i, cal in enumerate(cals):
            band = obs_info.det_info["band"][i]
            chan = obs_info.det_info["channel"][i]
            detset = obs_info.det_info["detset"][i]
            iva = ivas[detset]

            if iva is None:  # No IV analysis for this detset
                continue

            ridx = np.where((iva.bands == band) & (iva.channels == chan))[0]
            if not ridx:  # Channel doesn't exist in IV analysis
                continue

            ridx = ridx[0]
            cal.bg = iva.bgmap[ridx]
            cal.polarity = iva.polarity[ridx]
            cal.r_n = iva.R_n[ridx]  # type: ignore
            cal.p_sat = iva.p_sat[ridx]  # type: ignore

        obs_biases = dict(
            zip(obs_info.bias_lines, obs_info.biases[:, 0] * 2 * rtm_bit_to_volt)
        )
        bias_line_is_valid = {k: True for k in obs_biases.keys()}

        # check to see if biases have changed between bias steps and obs
        for bsa in bsas.values():
            if bsa is None:
                continue

            for bg, vb_bsa in enumerate(bsa.Vbias):
                bl_label = f"{bsa.meta['stream_id']}_b{bg:0>2}"
                # Usually we can count on bias voltages of bias lines >= 12 to be
                # Nan, however we have seen cases where they're not, so we also
                # restrict by count.
                if np.isnan(vb_bsa) or bg >= TES_BIAS_COUNT:
                    bias_line_is_valid[bl_label] = False
                    continue

                if np.abs(vb_bsa - obs_biases[bl_label]) > 0.1:
                    logger.debug(
                        "bias step and obs biases don't match for %s", bl_label
                    )
                    bias_line_is_valid[bl_label] = False

        # Add TES corrected params
        correction_results: Dict[str, List[tpc.CorrectionResults]] = {}
        if cfg.apply_cal_correction:
            logger.debug("Applying TES param corrections (%s)", obs_id)
            for dset in bsas:
                # logger.debug(f"Applying correction for {dset}")
                rs = []
                if executor is None:
                    for b, c in zip(ivas[dset].bands, ivas[dset].channels):
                        chdata = tpc.RpFitChanData.from_data(
                            ivas[dset], bsas[dset], b, c
                        )
                        rs.append(
                            tpc.run_correction(chdata, cfg.param_correction_config)
                        )
                else:
                    rs = tpc.run_corrections_parallel(
                        ivas[dset], bsas[dset], cfg.param_correction_config, executor=executor,
                        as_completed_callable=as_completed_callable)
                correction_results[dset] = rs
        res.correction_results = correction_results

        def find_correction_results(band, chan, dset):
            for r in correction_results[dset]:
                if r.chdata.band == band and r.chdata.channel == chan:
                    return r
            return None

        for i, cal in enumerate(cals):
            band = obs_info.det_info["band"][i]
            chan = obs_info.det_info["channel"][i]
            detset = obs_info.det_info["detset"][i]
            stream_id = obs_info.det_info["stream_id"][i]
            bg = cal.bg
            bsa = bsas[detset]

            if bsa is None or bg == -1:
                continue

            bl_label = f"{stream_id}_b{bg:0>2}"
            if not bias_line_is_valid[bl_label]:
                continue

            ridx = np.where((bsa.bands == band) & (bsa.channels == chan))[0]
            if not ridx:  # Channel doesn't exist in bias step analysis
                continue

            if cfg.apply_cal_correction:
                correction = find_correction_results(band, chan, detset)
                if correction is None:
                    logger.warn(
                        "Unable to find correction result for %s %s %s (%s)",
                        band,
                        chan,
                        detset,
                        obs_id,
                    )
                    use_correction = False
                    cal.tes_param_correction_success = False
                else:
                    use_correction = correction.success
                    cal.tes_param_correction_success = correction.success
            else:
                use_correction = False

            ridx = ridx[0]
            cal.tau_eff = bsa.tau_eff[ridx]
            if bg != -1:
                cal.v_bias = bsa.Vbias[bg]

            if use_correction and correction.corrected_params is not None:
                cpars = correction.corrected_params
                cal.r_tes = cpars.corrected_R0
                cal.r_frac = cpars.corrected_R0 / cal.r_n
                cal.s_i = cpars.corrected_Si * 1e6
                cal.p_bias = cpars.corrected_Pj * 1e-12
                cal.loopgain = cpars.loopgain
            else:
                cal.r_tes = bsa.R0[ridx]
                cal.r_frac = bsa.Rfrac[ridx]
                cal.p_bias = bsa.Pj[ridx]
                cal.s_i = bsa.Si[ridx]

            # Save naive parameters even if we're using corrected version
            cal.naive_r_tes = bsa.R0[ridx]
            cal.naive_r_frac = bsa.Rfrac[ridx]
            cal.naive_s_i = bsa.Si[ridx]
            cal.naive_p_bias = bsa.Pj[ridx]

            if cal.s_i == 0:
                cal.phase_to_pW = np.nan
            else:
                cal.phase_to_pW = pA_per_phi0 / (2 * np.pi) / cal.s_i * cal.polarity

            # Add bandpass informaton from bias group
            tube_flavor = obs_info.tube_flavor
            if cal.bg in BGS['lb']:
                cal.bandpass = BAND_STR[tube_flavor]['lb']
            elif cal.bg in BGS['hb']:
                cal.bandpass = BAND_STR[tube_flavor]['hb']

        res.result_set = np.array([astuple(c) for c in cals], dtype=CalInfo.dtype())
        res.success = True
    except Exception as e:
        res.traceback = traceback.format_exc()
        res.fail_msg = res.traceback
        if cfg.raise_exceptions:
            raise e
    return res


def get_obsids_to_run(cfg: DetCalCfg) -> List[str]:
    """
    Returns list of obs-ids to process, based on the configuration object.
    This will included non-processed obs-ids that are not found in the fail cache,
    and will be limitted to cfg.num_obs.
    """
    ctx = core.Context(cfg.context_path, metadata_list=cfg.metadata_list)
    # Find all obs_ids that have not been processed
    with open(cfg.failed_cache_file, "r") as f:
        failed_cache = yaml.safe_load(f)

    if failed_cache is not None:
        failed_obsids = set(failed_cache.keys())
    else:
        failed_obsids = set()

    db = core.metadata.ManifestDb(cfg.index_path)
    obs_ids_all = set(ctx.obsdb.query('type=="obs"')["obs_id"])
    processed_obsids = set(db.get_entries(["dataset"])["dataset"])
    obs_ids = sorted(list(obs_ids_all - processed_obsids - failed_obsids), reverse=True)
    if cfg.num_obs is not None:
        obs_ids = obs_ids[: cfg.num_obs]
    return obs_ids


def add_to_failed_cache(cfg: DetCalCfg, obs_id: str, msg: str) -> None:
    if "KeyboardInterrupt" in msg:  # Don't cache keyboard interrupts
        return

    if cfg.cache_failed_obsids:
        logger.info(f"Adding {obs_id} to failed_file_cache")
        with open(cfg.failed_cache_file, "r") as f:
            d = yaml.safe_load(f)
        if d is None:
            d = {}
        d[str(obs_id)] = msg
        with open(cfg.failed_cache_file, "w") as f:
            yaml.dump(d, f)
        return


def handle_result(result: CalRessetResult, cfg: DetCalCfg) -> None:
    """
    Handles a CalRessetResult. If successful, this will add to the manifestdb,
    if not this will add to the failed cache if cfg.cache_failed_obsids is True.
    """
    obs_id = str(result.obs_info.obs_id)
    if not result.success:
        logger.error(f"Failed on obs_id: {obs_id}")
        logger.error(result.traceback)

        msg = result.fail_msg
        if msg is None:
            msg = "unknown error"
        if 'sotodlib.core.metadata.loader.LoaderError' in msg:
            logger.error(f"obs_id {obs_id} failed due to medatada loader "
                         "error, try again later")
            return
        add_to_failed_cache(cfg, obs_id, msg)
        return

    logger.info(f"Adding obs_id {obs_id} to dataset")
    rset = ResultSet.from_friend(result.result_set)
    h5_path = cfg.h5_path
    if cfg.h5_unix_digits:
        name, ext = os.path.splitext(cfg.h5_path)
        unixtime = obs_id.split('_')[1][:cfg.h5_unix_digits]
        h5_path = f"{name}_{unixtime}{ext}"
    write_dataset(rset, h5_path, obs_id, overwrite=True)
    db = core.metadata.ManifestDb(cfg.index_path)
    relpath = os.path.relpath(h5_path, start=os.path.dirname(cfg.index_path))
    db.add_entry(
        {"obs:obs_id": obs_id, "dataset": obs_id}, filename=relpath, replace=True
    )


def run_update_site(
        cfg: DetCalCfg,
        executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
        as_completed_callable: Callable) -> None:
    """
    Main run script for computing det-cal results at the site. This will
    loop over obs-ids and serially gather the ObsInfo from the filesystem and
    sqlite dbs, and then compute the calibration results. A processing pool
    of cfg.nprocs_result_set processes will be used to parallelize the TES
    correction computation. If you have lots of compute power, and are limitted
    by filesystem or sqlite access, consider using the 'nersc' update function.

    Args:
    ------
    cfg: DetCalCfg or str
        DetCalCfg object or path to config yaml file
    """
    logger.setLevel(getattr(logging, cfg.log_level.upper()))
    for ch in logger.handlers:
        ch.setLevel(getattr(logging, cfg.log_level.upper()))

    obs_ids = get_obsids_to_run(cfg)

    logger.info(f"Processing {len(obs_ids)} obsids...")

    for oid in tqdm(obs_ids, disable=(not cfg.show_pb)):
        res = get_obs_info(cfg, oid)
        if not res.success:
            logger.info(f"Could not get obs info for obs id: {oid}")
            logger.error(res.traceback)

        if res.obs_info is None:
            continue

        result_set = get_cal_resset(cfg, res.obs_info, executor=executor,
                                    as_completed_callable=as_completed_callable)
        handle_result(result_set, cfg)


def run_update_nersc(cfg: DetCalCfg) -> None:
    """
    Main run script for computing det-cal results. This does the same thing as
    ``run_update_site`` however instantiates two separate pools for gathering
    ObsInfo and computing the ResultSets. This is useful in situations where
    sqlite/filesystem access are bottlenecks (such as nersc) so that ObsInfo can
    be gathered in parallel, and this can be done while ResultSet computation is
    ongoing. Because concurrent sqlite access can be limitted, it is recommended
    to keep cfg.nprocs_obs_info low (<10), while ``cfg.nprocs_result_set`` can be
    set arbitrarily large as to use remaining available resources.

    Args:
    ------
    cfg: DetCalCfg or str
        DetCalCfg object or path to config yaml file
    """
    logger.setLevel(getattr(logging, cfg.log_level.upper()))
    for ch in logger.handlers:
        ch.setLevel(getattr(logging, cfg.log_level.upper()))

    obs_ids = get_obsids_to_run(cfg)
    # obs_ids = ['obs_1713962395_satp1_0000100']
    # obs_ids = ['obs_1713758716_satp1_1000000']
    # obs_ids = ['obs_1701383445_satp3_1000000']
    logger.info(f"Processing {len(obs_ids)} obsids...")

    pb = tqdm(total=len(obs_ids), disable=(not cfg.show_pb))

    def callback(fut: Future):
        result = fut.result()
        pb.update()
        handle_result(result, cfg)

    def errback(e):
        logger.info(e)
        raise e

    # We split into multiple pools because:
    # - we don't want to overload sqlite files with too much concurrent access
    # - we want to be able to continue getting the next obs_info data while
    # ressets are being computed
    executor1 = ProcessPoolExecutor(max_workers=cfg.nprocs_obs_info)
    executor2 = ProcessPoolExecutor(max_workers=cfg.nprocs_result_set)

    resset_futures: list[Future] = []
    obsinfo_futures: list[Future] = []


    def get_obs_info_callback(fut: Future):
        try:
            result = fut.result()
            if result.success:
                future = executor2.submit(get_cal_resset, cfg, result.obs_info)
                future.add_done_callback(callback)
                resset_futures.append(future)
            else:
                pb.update()
                add_to_failed_cache(cfg, result.obs_id, result.traceback)
                logger.error(f"Failed to get obs_info for {result.obs_id}:\n{result.traceback}")
        except Exception as e:
            errback(e)

    try:
        for obs_id in obs_ids:
            future = executor1.submit(get_obs_info, cfg, obs_id)
            future.add_done_callback(get_obs_info_callback)
            obsinfo_futures.append(future)

        # Wait for all obsinfo tasks to complete
        for fut in as_completed(obsinfo_futures):
            pass  # results handled in callback

        # Wait for all resset tasks to complete
        for fut in as_completed(resset_futures):
            try:
                fut.result()  # Force exceptions to be raised here if any
            except Exception as e:
                errback(e)

    finally:
        executor1.shutdown(wait=True, cancel_futures=True)
        executor2.shutdown(wait=True, cancel_futures=True)

    pb.close()
    logger.info("Finished updates")


def get_parser(
    parser: Optional[argparse.ArgumentParser] = None,) -> argparse.ArgumentParser:
    if parser is None:
        p = argparse.ArgumentParser()
    else:
        p = parser
    p.add_argument(
        "config_file", type=str, help="yaml file with configuration for update script."
    )
    return p


def _main(
    cfg: DetCalCfg,
    executor: Union["MPICommExecutor", "ProcessPoolExecutor"],
    as_completed_callable: Callable) -> None:
    """
    Run update function. This will chose the correct method to run based on
    ``cfg.run_method``.
    """

    if cfg.run_method == "site":
        run_update_site(cfg, executor, as_completed_callable)
    elif cfg.run_method == "nersc":
        # Instantiate our own separate pools when running on NERSC.
        executor.shutdown(wait=True)
        run_update_nersc(cfg)
    else:
        raise ValueError(f"Unknown run_method: {cfg.run_method}")


def main(config_file: str):
    cfg = DetCalCfg.from_yaml(config_file)
    rank, executor, as_completed_callable = get_exec_env(cfg.nprocs_result_set)
    if rank == 0:
        _main(cfg, executor, as_completed_callable)

if __name__ == "__main__":
    sp_util.main_launcher(main, get_parser)



import datetime
from sotodlib import core
import logging
import numpy as np

logger = logging.getLogger(__name__)


def correct_iir_params(aman, ignore_time=False, check_srate=-1):
    """
    Correct missing iir_params by default values.
    This corrects iir_params only when the observation is within the time_range
    that is known to have problem.

    See `sotodlib.tod_ops.filters.iir_filter` for more details of iir_params

    Parameters
    ----------
    aman: AxisManager of observation

    ignore_time: Boolean. True if we don't want to check if the observation is within
                 a known bad time range.

    check_srate: If greater than 0 will check that the observations sample rate is within
                 check_srate Hz of 200 Hz. If less than 0 the check is skipped.

    Returns
    -------
    List of field names that have no iir_params

    """
    a = np.array([1., -3.74145562, 5.25726624, -3.28776591,
                  0.77203984, 0., 0., 0.,
                  0., 0., 0., 0.,
                  0., 0., 0., 0.])
    b = np.array([5.28396689e-06, 2.11358676e-05, 3.17038014e-05, 2.11358676e-05,
                  5.28396689e-06, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                  0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00,
                  0.00000000e+00, 0.00000000e+00, 0.00000000e+00, 0.00000000e+00])

    time_ranges = {
        'satp1': None,
        'satp2': None,
        'satp3': [('2023-01-01', '2024-10-24')],
        'lat': [(1745970733, 1746624934)],
    }[aman.obs_info.telescope]

    iir_missing = []
    for _field, _sub_iir_params in aman.iir_params._fields.items():
        if isinstance(_sub_iir_params, core.AxisManager):
            if 'a' in _sub_iir_params._fields:
                if _sub_iir_params['a'] is None:
                    iir_missing.append(_field)

    within_range = False
    if time_ranges is not None and not ignore_time:
        for t0, t1 in time_ranges:
            if isinstance(t0, str):
                t0 = datetime.datetime.strptime(t0, '%Y-%m-%d')
                t0 = t0.replace(tzinfo=datetime.timezone.utc).timestamp()
            if isinstance(t1, str):
                t1 = datetime.datetime.strptime(t1, '%Y-%m-%d')
                t1 = t1.replace(tzinfo=datetime.timezone.utc).timestamp()
            within_range = aman.timestamps[0] >= t0 and aman.timestamps[-1] <= t1
            if within_range:
                break

    if check_srate >= 0:
        srate = 1./np.mean(np.diff(aman.timestamps))
        if abs(srate - 200) > check_srate:
            raise ValueError(f"Sample rate is {srate}, too far from 200 Hz to use default params.")

    if within_range or ignore_time:
        for field in iir_missing:
            logger.warning(f'iir_params are missing on {field}. '
                           'Fill default params.')
            aman[f'iir_params.{field}.a'] = a
            aman[f'iir_params.{field}.b'] = b
    else:
        if len(iir_missing) > 0:
            raise ValueError('iir_params are missing but the observaiton is '
                             'not in a time range that is known to have a '
                             'problem. ' + ' '.join(iir_missing))
    return iir_missing



# Copyright (c) 2020-2024 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.

import os

import traitlets
import numpy as np

from toast.utils import Logger
from toast.mpi import MPI
import toast.qarray as qa
from toast.traits import trait_docs, Int, Unicode, Bool, Float, Instance
from toast.timing import function_timer, Timer
from toast.covariance import covariance_invert, covariance_apply
from toast.observation import default_values as defaults
from toast.ops import (
    Operator, BuildPixelDistribution, BuildInverseCovariance, BuildNoiseWeighted
)


@trait_docs
class Hn(Operator):
    """Evaluate geometrical h_n factors to support map-based simulations.

    """

    # Class traits

    API = Int(0, help="Internal interface version for this operator")

    pixel_pointing = Instance(
        klass=Operator,
        allow_none=True,
        help="This must be an instance of a pixel pointing operator",
    )

    hwp_angle = Unicode(
        None, allow_none=True, help="Observation shared key for HWP angle"
    )

    pixel_dist = Unicode(
        "pixel_dist",
        help="The Data key where the PixelDist object should be stored",
    )

    det_flags = Unicode(
        defaults.det_flags,
        allow_none=True,
        help="Observation detdata key for flags to use",
    )

    det_flag_mask = Int(
        defaults.det_mask_invalid, help="Bit mask value for optional detector flagging"
    )

    shared_flags = Unicode(
        defaults.shared_flags,
        allow_none=True,
        help="Observation shared key for telescope flags to use",
    )

    shared_flag_mask = Int(
        defaults.shared_mask_invalid,
        help="Bit mask value for optional telescope flagging",
    )

    output_dir = Unicode(
        ".",
        help="Write output data products to this directory",
    )

    noise_model = Unicode(
        "noise_model",
        allow_None=True,
        help="Observation key containing the noise model",
    )

    rcond_threshold = Float(
        1.0e-8, help="Minimum value for inverse condition number cut."
    )

    sync_type = Unicode(
        "alltoallv", help="Communication algorithm: 'allreduce' or 'alltoallv'"
    )

    save_pointing = Bool(
        False, help="If True, do not clear detector pointing matrices after use"
    )

    include_pol_angle = Bool(
        False,
        help="If True, include the polarization angle in the angle used to compute h_n",
    )

    nmin = Int(1, help="Minimum `n` to evaluate.")
    nmax = Int(4, help="Maximum `n` to evaluate.")

    cos_1_name = "cos_1"
    sin_1_name = "sin_1"
    cos_n_name = "cos_n"
    sin_n_name = "sin_n"
    hweight_name = "hweight"
    h_n_map = "h_n_map"
    covariance = "h_n_cov"

    @traitlets.validate("nmin")
    def _check_min(self, proposal):
        nmin = proposal["value"]
        if nmin < 1:
            raise traitlets.TraitError("Nmin should be greater than 0")
        return nmin

    @traitlets.validate("nmax")
    def _check_max(self, proposal):
        nmax = proposal["value"]
        if nmax < 1:
            raise traitlets.TraitError("Nmax should be greater than 0")
        return nmax

    @traitlets.validate("pixel_pointing")
    def _check_pixel_pointing(self, proposal):
        pixels = proposal["value"]
        if pixels is not None:
            if not isinstance(pixels, Operator):
                raise traitlets.TraitError(
                    "pixel_pointing should be an Operator instance"
                )
            # Check that this operator has the traits we expect
            for trt in ["pixels", "create_dist", "view"]:
                if not pixels.has_trait(trt):
                    msg = f"pixel_pointing operator should have a '{trt}' trait"
                    raise traitlets.TraitError(msg)
        return pixels

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        return

    @function_timer
    def _get_h_n(self, data, n, det):
        """ Compute and store the next order of h_n
        """
        for obs in data.obs:
            if det not in obs.local_detectors:
                continue
            # HWP angle is not yet used but will be needed soon
            if self.hwp_angle is not None:
                hwpang = obs.shared[self.hwp_angle]
            else:
                hwpang = None
            if n == 1:
                for name in [
                    self.cos_1_name,
                    self.sin_1_name,
                    self.hweight_name,
                    self.cos_n_name,
                    self.sin_n_name,
                ]:
                    obs.detdata.ensure(name, detectors=[det])
                
                # Compute detector quaternions
                obs_data = data.select(obs_uid=obs.uid)
                self.pixel_pointing.detector_pointing.apply(obs_data, detectors=[det])
                quats = obs.detdata[self.pixel_pointing.detector_pointing.quats][det]
                theta, phi, psi = qa.to_iso_angles(quats)
                if not self.include_pol_angle:
                    # FIXME: temporary hack until instrument classes are also pre-staged
                    # to GPU -- taken from derivatives_and_beams toast3 branch 
                    focalplane = obs.telescope.focalplane
                    focalplane.detector_data
                    psi -= focalplane[det]["pol_angle"].value # Removing the polarization angle
                cos_n_new = np.cos(psi)
                sin_n_new = np.sin(psi)
                obs.detdata[self.cos_1_name][det] = cos_n_new
                obs.detdata[self.sin_1_name][det] = sin_n_new
                obs.detdata[self.hweight_name][det] = 1
            else:
                # Use the angle sum identities to evaluate the
                # next cos(n * psi) and sin(n * psi)
                cos_1 = obs.detdata[self.cos_1_name][det]
                sin_1 = obs.detdata[self.sin_1_name][det]
                cos_n_old = obs.detdata[self.cos_n_name][det].copy()
                sin_n_old = obs.detdata[self.sin_n_name][det].copy()
                cos_n_new = cos_n_old * cos_1 - sin_n_old * sin_1
                sin_n_new = sin_n_old * cos_1 + cos_n_old * sin_1
            obs.detdata[self.cos_n_name][det] = cos_n_new
            obs.detdata[self.sin_n_name][det] = sin_n_new
        return

    @function_timer
    def _get_covariance(self, data, det):
        self.pixel_pointing.apply(data, detectors=[det])

        BuildInverseCovariance(
            pixel_dist=self.pixel_dist,
            inverse_covariance=self.covariance,
            view=self.pixel_pointing.view,
            det_flags=self.det_flags,
            det_flag_mask=self.det_flag_mask,
            shared_flags=self.shared_flags,
            shared_flag_mask=self.shared_flag_mask,
            pixels=self.pixel_pointing.pixels,
            weights=self.hweight_name,
            noise_model=self.noise_model,
            sync_type=self.sync_type,
        ).apply(data, detectors=[det])

        covariance_invert(
            data[self.covariance],
            self.rcond_threshold,
            use_alltoallv=(self.sync_type == "alltoallv"),
        )
        return


    @function_timer
    def _save_h_n(self, data, n, det):
        """ Accumulate and save the next order of h_n
        """
        if n < self.nmin:
            return

        log = Logger.get()

        for name, det_data in ("cos", self.cos_n_name), ("sin", self.sin_n_name):
            build_zmap = BuildNoiseWeighted(
                pixel_dist=self.pixel_dist,
                zmap=self.h_n_map,
                view=self.pixel_pointing.view,
                pixels=self.pixel_pointing.pixels,
                weights=self.hweight_name,
                noise_model=self.noise_model,
                det_data=det_data,
                det_flags=self.det_flags,
                det_flag_mask=self.det_flag_mask,
                shared_flags=self.shared_flags,
                shared_flag_mask=self.shared_flag_mask,
                sync_type=self.sync_type,
            )
            build_zmap.apply(data, detectors=[det])

            covariance_apply(
                data[self.covariance],
                data[self.h_n_map],
                use_alltoallv=(self.sync_type == "alltoallv"),
            )

            fname = os.path.join(self.output_dir, f"{self.name}_{det}_{name}_{n}.fits")
            data[self.h_n_map].write(fname)
            log.info_rank(f"Wrote h_n map to {fname}", comm=data.comm.comm_world)

        return

    @function_timer
    def _get_detectors(self, data):
        """ Find a super set of detectors across all processes in the
        communicator.
        """
        my_detectors = set(data.all_local_detectors())
        comm = data.comm.comm_world
        if comm is None:
            all_detectors = my_detectors
        else:
            all_detectors = comm.gather(my_detectors)
            if comm.rank == 0:
                for detectors in all_detectors:
                    my_detectors = my_detectors.union(detectors)
            all_detectors = comm.bcast(my_detectors)
        return all_detectors

    @function_timer
    def _exec(self, data, detectors=None, **kwargs):
        log = Logger.get()

        for trait in "pixel_pointing", "pixel_dist", "noise_model":
            if not hasattr(self, trait):
                msg = f"You must set the '{trait}' trait before calling exec()"
                raise RuntimeError(msg)

        if data.comm.world_rank == 0:
            os.makedirs(self.output_dir, exist_ok=True)

        # To accumulate, we need the pixel distribution.

        if self.pixel_dist not in data:
            pix_dist = BuildPixelDistribution(
                pixel_dist=self.pixel_dist,
                pixel_pointing=self.pixel_pointing,
                shared_flags=self.shared_flags,
                shared_flag_mask=self.shared_flag_mask,
                save_pointing=self.save_pointing,
            )
            log.info_rank("Caching pixel distribution", comm=data.comm.comm_world)
            pix_dist.apply(data)

        detectors = self._get_detectors(data)

        for det in detectors:
            for n in range(1, self.nmax + 1):
                self._get_h_n(data, n, det)
                self._get_covariance(data, det)
                self._save_h_n(data, n, det)

            # Purge the intermediate data objets

            for obs in data.obs:
                for name in (
                        self.hweight_name,
                        self.cos_1_name, self.sin_1_name,
                        self.cos_n_name, self.sin_n_name
                ):
                    del obs.detdata[name]

        return

    def _finalize(self, data, **kwargs):
        return

    def _requires(self):
        req = self.pixel_pointing.requires()
        req["meta"].extend([self.noise_model])
        req["detdata"].extend([self.det_data])
        if self.det_flags is not None:
            req["detdata"].append(self.det_flags)
        return req

    def _provides(self):
        prov = {"meta": [self.binned], "shared": list(), "detdata": list()}
        return prov

    def _accelerators(self):
        return list()



# Copyright (c) 2018-2020 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Hardware configuration utilities.
"""

import os
import re
import copy

from collections import OrderedDict

import gzip

import astropy.units as u
import numpy as np

import toml


# The extra rotation to apply to the projected LAT focalplane on the
# sky is given by:
#
#      R = Elevation - Offset - Corotator
#
# and the offset is defined as the "design elevation" which places
# the focalplane at the correct orientation with no co-rotation.

LAT_COROTATOR_OFFSET = u.Quantity(60.0, u.degree)


def build_readout_id(creation_time, wafer_slot, channel):
    """Construct a simulated readout_id.

    We build this string from the stream_id (wafer slot), the overall
    creation time (treating this as the last tuning time), and the
    readout channel.  We convert the readout channel into a smurf band
    and smurf channel assuming 8 bands and 512 channels per band.

    Args:
        creation_time (float):  The POSIX ctime of the simulation start.
        wafer_slot (str):  The stream_id / wafer_slot
        channel (int):  The readout channel (0-4095).

    Returns:
        (str):  The fake readout_id.

    """
    creation_time = int(creation_time)
    smurf_band = channel // 8
    smurf_channel = channel % 8
    return f"{wafer_slot}_{creation_time:10d}_{smurf_band}_{smurf_channel}"

def parse_readout_id(readout_id):
    """Split a readout_id into its parts.

    Args:
        creation_time (float):  The POSIX ctime of the simulation start.
        wafer_slot (str):  The stream_id / wafer_slot
        channel (int):  The readout channel (0-4095).

    Returns:
        (tuple):  The wafer_slot, creation time, readout channel

    """
    pat = re.compile(r"(.*)_(.*)_(.*)_(.*)")
    mat = pat.match(readout_id)
    if mat is None:
        raise ValueError(f"Readout ID {readout_id} is invalid")
    wf = mat.group(1)
    ct = float(mat.group(2))
    smband = int(mat.group(3))
    smchan = int(mat.group(4))
    channel = smband * 8 + smchan
    return (wf, ct, channel)

def sim_wafer_names( hw ):
    """Adds SO generic UFM names to the hardware model based on the type of the wafer
       Ex: Uv1, Mv4, Lv3, etc
    """
    c = [1,1,1]
    for wafer in hw.data["wafer_slots"]:
        wprops = hw.data["wafer_slots"][wafer]
        if "UHF" in wprops["type"]:
            pre = "Uv"
            i = 0
        elif "MF" in wprops["type"]:
            pre = "Mv"
            i = 1
        elif "LF" in wprops["type"]:
            pre = "Lv"
            i = 2
        else:
            raise ValueError(f"Unknown band type {wprops['type']} for wafer {wafer}")

        wprops["wafer_name"] = f"{pre}{c[i]}"
        c[i] += 1


class Hardware(object):
    """Class representing a specific hardware configuration.

    The data is stored in a dictionary, and can be loaded / dumped to disk
    as well as trimmed to include only a subset of detectors.

    Args:
        path (str, optional): If specified, configuration is loaded from this
            file during construction.

    """
    def __init__(self, path=None):
        self.data = OrderedDict()
        if path is not None:
            self.load(path)

    def dump(self, path, overwrite=False, compress=False):
        """Write hardware config to a TOML file.

        Dump data to a TOML format file, optionally compressing the contents
        with gzip and optionally overwriting the file.

        Args:
            path (str): The file to write.
            overwrite (bool): If True, overwrite the file if it exists.
                If False, then existing files will cause an exception.
            compress (bool): If True, compress the data with gzip on write.

        Returns:
            None

        """
        if os.path.exists(path):
            if overwrite:
                os.remove(path)
            else:
                raise RuntimeError("Dump path {} already exists.  Use "
                                   "overwrite option".format(path))
        if compress:
            with gzip.open(path, "wb") as f:
                dstr = toml.dumps(self.data)
                f.write(dstr.encode())
        else:
            with open(path, "w") as f:
                dstr = toml.dumps(self.data)
                f.write(dstr)
        return

    def load(self, path):
        """Read data from a TOML file.

        The file can either be regular text or a gzipped version of a TOML
        file.

        Args:
            path (str): The file to read.

        Returns:
            None

        """
        dstr = None
        try:
            with gzip.open(path, "rb") as f:
                dstr = f.read()
                self.data = toml.loads(dstr.decode())
        except OSError:
            with open(path, "r") as f:
                dstr = f.read()
                self.data = toml.loads(dstr)
        return

    def wafer_map(self):
        """Construct wafer mapping to other auxilliary data.

        Given the current data state, build dictionaries to go from wafer_slots
        to all other non-detector info:  telescopes, tube_slots, card_slots, crate_slots,
        and bands.  This is a convenient mapping when pruning the hardware
        information or doing other kinds of lookups.

        Returns:
            (dict): Nested dictionaries from wafers to other properties.

        """
        result = OrderedDict()

        tube_to_tele = dict()
        for tele, props in self.data["telescopes"].items():
            for tb in props["tube_slots"]:
                tube_to_tele[tb] = tele

        wafer_to_tube = dict()
        for tb, props in self.data["tube_slots"].items():
            for wf in props["wafer_slots"]:
                wafer_to_tube[wf] = tb

        crate_to_card = dict()
        for crate, props in self.data["crate_slots"].items():
            for card in props["card_slots"]:
                crate_to_card[card] = crate

        result["card_slots"] = {x: y["card_slot"]
                           for x, y in self.data["wafer_slots"].items()}
        result["crate_slots"] = {x: crate_to_card[y["card_slot"]]
                            for x, y in self.data["wafer_slots"].items()}
        result["bands"] = {x: y["bands"]
                           for x, y in self.data["wafer_slots"].items()}
        result["tube_slots"] = wafer_to_tube
        result["telescopes"] = {x: tube_to_tele[wafer_to_tube[x]] for x in
                                list(self.data["wafer_slots"].keys())}
        return result

    def select(self, telescopes=None, tube_slots=None, match=dict()):
        """Select a subset of detectors.

        Select detectors whose properties match some criteria.  A new Hardware
        object is created and returned.  If a matching expression is not
        specified for a given property name, then this is equivalent to
        selecting all values of that property.

        Before selecting on detector properties, any telescope / tube_slot filtering
        criteria are first applied.

        Each key of the "match" dictionary should be the name of a detector
        property to be considered for selection (e.g. band, wafer, pol, pixel).
        The value is a matching expression which can be:

            - A list of explicit values to match.
            - A string containing a regex expression to apply.

        Example:
            Imagine you wanted to select all 90GHz detectors on wafers 25 and
            26 which have "A" polarization and are located in pixels 20-29
            (recall the "." matches a single character)::

                new = hw.select(match={"wafer_slot": ["w25", "w26"],
                                "band": "SAT_f090",
                                "pol": "A",
                                "pixel": "02."})

        Args:
            telescopes (str): A regex string to apply to telescope names or a
                list of explicit names.
            tube_slots (str): A regex string to apply to tube_slot names or a list of
                explicit names.
            match (dict): The dictionary of property names and their matching
                expressions.

        Returns:
            (Hardware): A new Hardware instance with the selected detectors.

        """
        # First parse any telescope and tube_slot options into a list of wafers
        wselect = None
        tbselect = None
        if telescopes is not None:
            tbselect = list()
            for tele in telescopes:
                tbselect.extend(self.data["telescopes"][tele]["tube_slots"])
        if tube_slots is not None:
            if tbselect is None:
                tbselect = list()
            tbselect.extend(tube_slots)
        if tbselect is not None:
            wselect = list()
            for tb in tbselect:
                wselect.extend(self.data["tube_slots"][tb]["wafer_slots"])

        dets = self.data["detectors"]

        # Build regex matches for each property
        reg = dict()
        if "wafer_slot" in match:
            # Handle wafer case separately, since we need to merge any
            # match with our telescope / tube_slot selection of wafers above.
            k = "wafer_slot"
            v = match[k]
            if wselect is None:
                # Just the regular behavior
                if isinstance(v, list):
                    reg[k] = re.compile(r"("+"|".join(v)+r")")
                else:
                    reg[k] = re.compile(v)
            else:
                # Merge our selection
                wall = list(wselect)
                if isinstance(v, list):
                    wall.extend(v)
                else:
                    wall.append(v)
                reg[k] = re.compile(r"("+"|".join(wall)+r")")
        elif wselect is not None:
            # No pattern in the match dictionary, just our list from the
            # telescope / tube selection.
            reg["wafer_slot"] = re.compile(r"("+"|".join(wselect)+r")")

        for k, v in match.items():
            if (k == "wafer_slot"):
                # Already handled above
                continue
            else:
                if isinstance(v, list):
                    reg[k] = re.compile(r"("+"|".join(v)+r")")
                else:
                    reg[k] = re.compile(v)

        # Go through all detectors selecting things that match all fields
        newwafers = set()
        newdets = OrderedDict()
        for d, props in dets.items():
            keep = True
            for k, v in reg.items():
                if k in props:
                    test = v.match(props[k])
                    if test is None:
                        keep = False
                        break
            if keep:
                newwafers.add(props["wafer_slot"])
                newdets[d] = copy.deepcopy(props)

        # Now compute the reduced set of auxilliary data needed for these
        # detectors.
        wafermap = self.wafer_map()

        # Copy this data
        hw = Hardware()
        hw.data = OrderedDict()
        for k, v in wafermap.items():
            hw.data[k] = OrderedDict()
            tocopy = set()
            for wf in newwafers:
                if isinstance(v[wf], list):
                    for iv in v[wf]:
                        tocopy.add(iv)
                else:
                    tocopy.add(v[wf])
            for elem in tocopy:
                hw.data[k][elem] = copy.deepcopy(self.data[k][elem])

        # Copy over the wafer data
        hw.data["wafer_slots"] = OrderedDict()
        for wf in newwafers:
            hw.data["wafer_slots"][wf] = copy.deepcopy(self.data["wafer_slots"][wf])

        # And the detectors...
        hw.data["detectors"] = newdets

        return hw



from .detrend import detrend_tod
from .fft_ops import rfft
from .filters import fourier_filter, fft_trim
from .gapfill import \
    get_gap_fill, get_gap_fill_single, \
    get_gap_model, get_gap_model_single
from . import jumps
from . import pca
from .apodize import apodize_cosine
from .binning import bin_signal
from .sub_polyf import subscan_polyfilter
from .azss import get_azss
from .t2pleakage import get_t2p_coeffs, subtract_t2p
from . import deproject



import re
import numpy as np

from ..core import metadata


def _has_tag(root, keys):
    keys = keys.split(".")
    for key in keys:
        if key in root:
            root = root[key]
        else:
            return False
    return True

def _get_tag(root, keys, i):
    keys = keys.split(".")
    for key in keys:
        root = root[key]
    return root[i]


class QAMetric(object):
    """ Base class for quality assurance metrics to be recorded in Influx,
    derived from processed metadata files.

    The base class includes methods for recording a metric value to Influx,
    checking if a given obs_id has been recorded for this metric, and fetching
    a list of obs_id's that can be recorded.

    Metrics are labelled by the observation ID they describe, an Influx 'measurement'
    and 'field', and any number of additional tags. A measurement can have multiple
    fields associated with it, and here we've tried to organize the metrics so that
    a source of information (e.g. HWP angle solutions) is a measurement and relevant
    quantities describing it (e.g. HWP success, mean rate, etc) are fields.

    Subclasses should implement the `_process` and `_get_available_obs` methods to
    access the source of data they are going to tap, as well as define the
    `_influx_meas` and `_influx_field` attributes.
    """

    # these will be defined by child classes
    _influx_meas = None  # measurement to record to
    _influx_field = None  # the field to populate

    def __init__(self, context, monitor, log="qa_metrics_log"):
        """ A QA metric base class.

        Arguments
        ---------
        context : core.Context
            Context that includes all necessary metadata to generate metrics.
        monitor : site_pipeline.monitor.Monitor
            InfluxDB connection.
        log : str
            InfluxDB measurement that is used to log new entries.
        """

        self.context = context
        self.monitor = monitor
        self._influx_log = log

    def exists(self, obs_id, tags={}):
        """ Check if a metric exists for this obs_id in Influx.

        Arguments
        ---------
        obs_id : str
            The observation ID to check.
        tags : dict (optional)
            Further restrict to given tags.

        Returns
        -------
        exists : bool
            Whether it exists or not.
        """
        return self.monitor.check(self._influx_field, obs_id, tags, log=self._influx_log)

    def get_existing_obs(self):
        """ Get a list of observations already recorded to Influx.
        """
        # query influx log measurement for observations of this field
        res = self.monitor.client.query(
            f"select {self._influx_field}, observation from {self._influx_log}"
        ).get_points()
        return [r["observation"] for r in res]

    def process_and_record(self, obs_id, meta=None):
        """ Generate a metric for this obs_id and record it to InfluxDB.

        Arguments
        ---------
        obs_id : str
            The observation ID to process.
        meta : AxisManager (optional)
            The metadata for this observation ID. If not provided will read from file.
        """
        if meta is None:
            meta = self.context.get_meta(obs_id, ignore_missing=True)
        if meta.obs_info.obs_id != obs_id:
            raise Exception(f"Metadata does not correspond to obs_id {obs_id}.")
        line = self._process(meta)
        log_tags = {"observation": obs_id}  # used to identify this entry
        self.monitor.record(**line, log=self._influx_log, measurement=self._influx_meas, log_tags=log_tags)
        self.monitor.write()

    def get_new_obs(self):
        """ Get a list of available observations not yet recorded to InfluxDB.
        """
        # get a list of obs_id to update
        avail_obs = set(self._get_available_obs())
        exist_obs = set(self.get_existing_obs())
        return avail_obs - exist_obs

    def _process(self, meta):
        """ Implement this to process an actual metric."""
        raise NotImplementedError

    def _get_available_obs(self):
        """ Implement this for a given metric."""
        raise NotImplementedError


class PreprocessQA(QAMetric):
    """ A metric derived from a preprocesstod process.
    """

    _influx_meas = "preprocesstod"
    _process_args = {}

    def __init__(self, context, monitor, process_name, process_args={}, **kwargs):
        """ In addition to the context and monitor, pass the name of the
        preprocess process to record. It should have a `gen_metric` method
        implemented and `_influx_field` attribute.
        """
        self._process_args = process_args
        super().__init__(context, monitor, **kwargs)

        from sotodlib.preprocess import Pipeline
        proc = Pipeline.PIPELINE.get(process_name, None)
        if proc is None:
            raise Exception(f"No preprocess process with name {process_name}")
        self._pipe_proc = proc

        # get the field name from the process
        self._influx_field = self._pipe_proc._influx_field

    def _process(self, meta):
        return self._pipe_proc.gen_metric(meta, meta.preprocess, **self._process_args)

    def _get_available_obs(self):
        # find preprocess manifest file
        man_file = [p["db"] for p in self.context["metadata"] if re.match(".*preprocess", p.get("db", ""))]
        if len(man_file) == 0:
            raise Exception(f"No preprocess metadata block in context {self.context.filename}.")

        # load manifest and read available observations
        man_db = metadata.ManifestDb.from_file(man_file[0])
        return [o[0] for o in man_db.get_entries(["\"obs:obs_id\""]).asarray()]


# inherit from PreprocessQA to reuse available_obs method
class PreprocessValidDets(PreprocessQA):
    """A metric for the number of detectors deemed valid at the end of the preprocesstod
    processing. For each wafer slot and bandpass, the number of detectors for which the
    fraction of samples that were valid is greater than a configurable threshold is recorded.

    The config entry supports a `process_args` block where the following options can be
    specified:

    tags : dict
        The values are keys into `metadata.det_info` to record as tags with
        the Influx line. The keys are addded to the default list
        ["wafer_slot", "tel_tube", "bandpass"] with "bandpass" being taken
        from "wafer.bandpass" or "det_cal.bandpass" if the former isn't found.
    thresh : float
        The threshold for the fraction of valid samples above which a detector is
        deemed good (default 0.75)
    process_name : str
        The process from which to read the valid dataset. (default 'glitches')

    """

    _influx_meas = "preprocesstod"
    _influx_field = "num_valid_dets"

    def __init__(
        self,
        *args,
        process_args={},
        **kwargs
    ):
        # bypass the PreprocessQA __init__
        super(PreprocessQA, self).__init__(*args, **kwargs)
        # extract parameters
        self._tags = process_args.get("tags", {})
        self._thresh = process_args.get("thresh", 0.75)
        self._key = process_args.get("process_name", "glitches")

    def _process(self, meta):

        # add specified tags
        tag_keys = {
            "wafer_slot": "wafer_slot",
            "tel_tube": "tel_tube",
        }

        if _has_tag(meta.det_info, 'wafer.bandpass'):
            bandpasses = meta.det_info.wafer.bandpass
            tag_keys["bandpass"] = "wafer.bandpass"
        else:
            bandpasses = meta.det_info.det_cal.bandpass
            tag_keys["bandpass"] = "det_cal.bandpass"

        tag_keys.update(self._tags)

        # record one metric per wafer slot, per bandpass
        # extract these tags for the metric
        tags = []
        vals = []
        for bp in np.unique(bandpasses):
            for ws in np.unique(meta.det_info.wafer_slot):
                subset = np.where(
                    (meta.det_info.wafer_slot == ws) & (bandpasses == bp)
                )[0]

                if len(subset) > 0:
                    # Compute the number of samples that are valid
                    frac_valid = np.array([
                        np.dot(r.ranges(), [-1, 1]).sum() / len(subset)
                        for r in meta.preprocess[self._key].valid[subset]
                    ])

                    # Count detectors with fraction valid above threshold
                    n_good = (frac_valid > self._thresh).sum()

                    # get the tags for this wafer (all detectors in this subset share these)
                    tags_i = {
                        k: _get_tag(meta.det_info, i, subset[0]) for k, i in tag_keys.items() if _has_tag(meta.det_info, i)
                    }
                    tags_i["telescope"] = meta.obs_info.telescope

                    # add tags and values to respective lists in order
                    vals.append(n_good)
                    tags.append(tags_i)

        obs_time = [meta.obs_info.timestamp] * len(vals)
        return {
            "field": self._influx_field,
            "values": vals,
            "timestamps": obs_time,
            "tags": tags,
        }


# inherit from PreprocessQA to reuse available_obs method
class PreprocessArrayNET(PreprocessQA):
    """Generate a QA metric for array NET values for each wafer slot and bandpass.

    The config entry supports a `process_args` block where the following
    options can be specified:

    tags : dict
        The values are keys into `metadata.det_info` to record as tags with
        the Influx line. The keys are addded to the default list
        ["wafer_slot", "tel_tube", "bandpass"] with "bandpass" being taken
        from "wafer.bandpass" or "det_cal.bandpass" if the former isn't found.
    noise_aman : str
        The name of the axis manager that holds the white noise array.
        (default 'noise')

    """

    _influx_meas = "preprocesstod"
    _influx_field = "array_net"

    def __init__(
        self,
        *args,
        process_args={},
        **kwargs
    ):
        # bypass the PreprocessQA __init__
        super(PreprocessQA, self).__init__(*args, **kwargs)
        # extract parameters
        self._tags = process_args.get("tags", {})
        self._noise_aman = process_args.get("noise_aman", "noise")
        self._field_name = process_args.get("field_name", "")
        self._unit_factor = process_args.get("unit_factor", 1)

    def _process(self, meta):

        # record one metric per wafer_slot per bandpass
        # extract these tags for the metric
        tag_keys = {
            "wafer_slot": "wafer_slot",
            "tel_tube": "tel_tube",
        }

        if _has_tag(meta.det_info, 'wafer.bandpass'):
            bandpasses = meta.det_info.wafer.bandpass
            tag_keys["bandpass"] = "wafer.bandpass"
        else:
            bandpasses = meta.det_info.det_cal.bandpass
            tag_keys["bandpass"] = "det_cal.bandpass"

        tag_keys.update(self._tags)
        tags = []
        vals = []
        for bp in np.unique(bandpasses):
            for ws in np.unique(meta.det_info.wafer_slot):
                subset = np.where(
                    (meta.det_info.wafer_slot == ws) & (bandpasses == bp)
                )[0]

                white_noise = meta.preprocess[self._noise_aman].white_noise[subset] * self._unit_factor
                mask = (white_noise != 0) & (~np.isnan(white_noise))
                good_indices = np.nonzero(mask)[0]
                if good_indices.size > 0:
                    vals.append(np.sqrt(1.0 / np.nansum(1.0 / (white_noise[good_indices])**2)))

                    tags_base = {
                        k: _get_tag(meta.det_info, i, subset[0]) for k, i in tag_keys.items() if _has_tag(meta.det_info, i)
                    }
                    tags_base["telescope"] = meta.obs_info.telescope
                    tags.append(tags_base)

        obs_time = [meta.obs_info.timestamp] * len(tags)
        return {
            "field": f"{self._influx_field}{'_' + self._field_name if self._field_name else ''}",
            "values": vals,
            "timestamps": obs_time,
            "tags": tags,
        }

# inherit from PreprocessQA to reuse available_obs method
class PreprocessDetNET(PreprocessQA):
    """Generate a QA metric for per detector NET values for each wafer slot
    and bandpass.

    The config entry supports a `process_args` block where the following
    options can be specified:

    tags : dict
        The values are keys into `metadata.det_info` to record as tags with
        the Influx line. The keys are addded to the default list
        ["wafer_slot", "tel_tube", "bandpass"] with "bandpass" being taken
        from "wafer.bandpass" or "det_cal.bandpass" if the former isn't found.
    noise_aman : str
        The name of the axis manager that holds the white noise array.
        (default 'noise')

    """

    _influx_meas = "preprocesstod"
    _influx_field = "det_net"

    def __init__(
        self,
        *args,
        process_args={},
        **kwargs
    ):
        # bypass the PreprocessQA __init__
        super(PreprocessQA, self).__init__(*args, **kwargs)
        # extract parameters
        self._tags = process_args.get("tags", {})
        self._noise_aman = process_args.get("noise_aman", "noise")
        self._field_name = process_args.get("field_name", "")
        self._unit_factor = process_args.get("unit_factor", 1)

    def _process(self, meta):

        # record one metric per wafer_slot per bandpass
        # extract these tags for the metric
        tag_keys = {
            "wafer_slot": "wafer_slot",
            "tel_tube": "tel_tube",
        }

        if _has_tag(meta.det_info, 'wafer.bandpass'):
            bandpasses = meta.det_info.wafer.bandpass
            tag_keys["bandpass"] = "wafer.bandpass"
        else:
            bandpasses = meta.det_info.det_cal.bandpass
            tag_keys["bandpass"] = "det_cal.bandpass"

        tag_keys.update(self._tags)
        tags = []
        vals = []
        for bp in np.unique(bandpasses):
            for ws in np.unique(meta.det_info.wafer_slot):
                subset = np.where(
                    (meta.det_info.wafer_slot == ws) & (bandpasses == bp)
                )[0]

                white_noise = meta.preprocess[self._noise_aman].white_noise[subset] * self._unit_factor
                mask = (white_noise != 0) & (~np.isnan(white_noise))
                good_indices = np.nonzero(mask)[0]
                if good_indices.size > 0:
                    vals.append(np.sqrt(1.0 / np.nansum(1.0 / (white_noise[good_indices])**2)) * np.sqrt(len(good_indices)))

                    tags_base = {
                        k: _get_tag(meta.det_info, i, subset[0]) for k, i in tag_keys.items() if _has_tag(meta.det_info, i)
                    }
                    tags_base["telescope"] = meta.obs_info.telescope
                    tags.append(tags_base)

        obs_time = [meta.obs_info.timestamp] * len(tags)
        return {
            "field": f"{self._influx_field}{'_' + self._field_name if self._field_name else ''}",
            "values": vals,
            "timestamps": obs_time,
            "tags": tags,
        }


class HWPSolQA(QAMetric):
    """ Base class for metrics derived from HWP angle solutions. Subclasses should
    implement the `_process` method. Some quantities are derived twice, once for each
    encoder, and these will require and `encoder` parameter to be provided to select
    which one to produce a metric for. This is indicated by setting the `_needs_encoder`
    attribute to `True`.
    """

    _influx_meas = "hwp_solution"
    _needs_encoder = False  # set this flag if encoder needs to be specified

    def __init__(self, context, monitor, encoder=None, **kwargs):
        super().__init__(context, monitor, **kwargs)

        self._encoder = encoder
        self._tags = {}
        if encoder is not None:
            if str(encoder) not in ["1", "2"]:
                raise Exception(f"Invalid value {encoder} for encoder parameter.")
            self._tags = {"encoder": self._encoder}
        elif self._needs_encoder:
            raise Exception("This metric needs an encoder to be specified on creation.")

    def _get_available_obs(self):
        # find preprocess manifest file
        man_file = [p["db"] for p in self.context["metadata"] if p.get("label", "") == "hwp_solution"]
        if len(man_file) == 0:
            raise Exception(f"No hwp_solution metadata block in context {self.context.filename}.")

        # load manifest and read available observations
        man_db = metadata.ManifestDb.from_file(man_file[0])
        obs_re = re.compile("^obs_*.")
        return [o[0] for o in man_db.get_entries(["\"obs:obs_id\""]).asarray() if obs_re.match(o[0])]

    def _process(self, meta):
        tags = self._tags.copy()
        tags.update({"telescope": meta.obs_info.telescope})
        obs_time = [meta.obs_info.timestamp]
        return {
            "field": self._influx_field,
            "values": [self._gen_value(meta)],
            "timestamps": obs_time,
            "tags": [tags],
        }


class HWPSolSuccess(HWPSolQA):
    """ Records success of the HWP angle solution calculation, for each encoder.
    0: No data, 1: Success, 2: Fail
    """

    _influx_field = "sol_success"
    _needs_encoder = True

    def _gen_value(self, meta):
        if meta.hwp_solution[f"logger_{self._encoder}"] == 'No HWP data':
            return 0
        elif meta.hwp_solution[f"logger_{self._encoder}"] == "Angle calculation succeeded":
            return 1
        else:
            return 2


class HWPSolPrimaryEncoder(HWPSolQA):
    """ The primary encoder used for the HWP angle calculation."""

    _influx_field = "primary_encoder"

    def _gen_value(self, meta):
        return meta.hwp_solution["primary_encoder"]


class HWPSolVersion(HWPSolQA):
    """ The version of the solution used for the HWP angle calculation."""

    _influx_field = "version"

    def _gen_value(self, meta):
        return meta.hwp_solution["version"]


class HWPSolOffcenter(HWPSolQA):
    """ Calculated offcentering of HWP angle solution."""

    _influx_field = "offcenter"

    def _gen_value(self, meta):
        return meta.hwp_solution["offcenter"][0]


class HWPSolOffcenterErr(HWPSolQA):
    """ Standard error on the offcentering of HWP angle solution."""

    _influx_field = "offcenter_err"

    def _gen_value(self, meta):
        return meta.hwp_solution["offcenter"][1]


class HWPSolNumSamples(HWPSolQA):
    """ The total number of encoder samples."""

    _influx_field = "num_samples"
    _needs_encoder = True

    def _gen_value(self, meta):
        flag_key = f"filled_flag_{self._encoder}"
        return meta.hwp_solution[flag_key].size


class HWPSolNumFlagged(HWPSolQA):
    """ The number of encoder samples that were flagged."""

    _influx_field = "num_flagged"
    _needs_encoder = True

    def _gen_value(self, meta):
        flag_key = f"filled_flag_{self._encoder}"
        return meta.hwp_solution[flag_key].sum()


class HWPSolMeanRate(HWPSolQA):
    """ The mean calculated HWP angle rate, for each encoder."""

    _influx_field = "mean_rate"
    _needs_encoder = True

    def _gen_value(self, meta):
        good_samp = ~meta.hwp_solution[f"filled_flag_{self._encoder}"]
        nsamp = good_samp.sum()
        rate = np.nan if nsamp == 0 else (meta.hwp_solution[f"hwp_rate_{self._encoder}"] * good_samp).sum() / nsamp
        return rate


class HWPSolMeanTemplate(HWPSolQA):
    """ The mean of the calculated template magnitude."""

    _influx_field = "mean_template"
    _needs_encoder = True

    def _gen_value(self, meta):
        return np.mean(np.abs(meta.hwp_solution[f"template_{self._encoder}"]))


class HWPSolNumDroppedPacketsEncoder(HWPSolQA):
    """ The number of dropped encoder packets."""

    _influx_field = "num_dropped_packets_encoder"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_dropped_packets_{self._encoder}"]


class HWPSolNumDroppedPacketsIrig(HWPSolQA):
    """ The number of dropped IRIG packets."""

    _influx_field = "num_dropped_packets_irig"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_dropped_packets_irig_{self._encoder}"]


class HWPSolNumDataPointGlitchesEncoder(HWPSolQA):
    """ The number of data point glitches in encoder data."""

    _influx_field = "num_data_point_glitches_encoder"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_glitches_{self._encoder}"]


class HWPSolNumDataPointGlitchesIrig(HWPSolQA):
    """ The number of data point glitches in IRIG data."""

    _influx_field = "num_data_point_glitches_irig"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_glitches_irig_{self._encoder}"]


class HWPSolNumValueGlitchesEncoder(HWPSolQA):
    """ The number of value glitches in encoder data."""

    _influx_field = "num_value_glitches_encoder"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_value_glitches_{self._encoder}"]


class HWPSolNumValueGlitchesIrig(HWPSolQA):
    """ The number of value glitches in IRIG data."""

    _influx_field = "num_value_glitches_irig"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_value_glitches_irig_{self._encoder}"]


class HWPSolDeadRotations(HWPSolQA):
    """ The number of dead rotations that failed to correct glitches."""

    _influx_field = "num_dead_rots"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_dead_rots_{self._encoder}"]


class HWPSolDroppedSlits(HWPSolQA):
    """ The number of dropped slits."""

    _influx_field = "num_dropped_slits"
    _needs_encoder = True

    def _gen_value(self, meta):
        return meta.hwp_solution[f"num_dropped_slits_{self._encoder}"]



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Test toast context loading.
"""

import os
import shutil
import glob
import re
import yaml

import numpy as np
import astropy.units as u

import unittest
from unittest import TestCase

# Import so3g before any other packages that import spt3g
import so3g

from ._helpers import create_outdir, simulation_test_data, close_data_and_comm


try:
    # Import sotodlib toast module first, which sets global toast defaults
    import sotodlib.toast as sotoast
    import sotodlib.toast.ops as so_ops
    import toast
    from toast.observation import default_values as defaults

    toast_available = True
except ImportError:
    raise
    toast_available = False

import sotodlib


class ToastContextTest(TestCase):
    def setUp(self):
        fixture_name = os.path.splitext(os.path.basename(__file__))[0]
        if not toast_available:
            print("TOAST cannot be imported- skipping unit tests", flush=True)
            return
        world, procs, rank = toast.get_world()
        self.outdir = create_outdir(fixture_name, mpicomm=world)

    def test_context_load(self):
        if not toast_available:
            return
        world, procs, rank = toast.get_world()

        cloader = so_ops.LoadContext(
            telescope_name="SAT1",
            context=None,
            readout_ids=['w25_p000_SAT_f090_A'],
        )


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2023-2023 Simons Observatory.
# Full license can be found in the top level "LICENSE" file.
"""Simple nominal noise models.
"""

import numpy as np
from astropy import units as u
import toast
import toast.ops

from .. import ops as so_ops
from .job import workflow_timer


def setup_simple_noise_models(operators):
    """Add commandline args and operators for simple noise models.

    These noise models use just nominal values from the focalplane
    and an elevation-weighting of the NET.

    Args:
        operators (list):  The list of operators to extend.

    Returns:
        None

    """
    operators.append(
        toast.ops.DefaultNoiseModel(name="default_model", noise_model="noise_model")
    )
    operators.append(
        toast.ops.ElevationNoise(name="elevation_model", out_model="noise_model")
    )
    operators.append(toast.ops.CommonModeNoise(name="common_mode_noise", enabled=False))


@workflow_timer
def simple_noise_models(job, otherargs, runargs, data):
    """Generate trivial noise models.

    Args:
        job (namespace):  The configured operators and templates for this job.
        otherargs (namespace):  Other commandline arguments.
        runargs (namespace):  Job related runtime parameters.
        data (Data):  The data container.

    Returns:
        None

    """
    log = toast.utils.Logger.get()
    timer = toast.timing.Timer()
    timer.start()

    # Configured operators for this job
    job_ops = job.operators

    # Default model from focalplane nominal values
    if job_ops.default_model.enabled:
        log.info_rank("  Running nominal noise model...", comm=data.comm.comm_world)
        job_ops.default_model.apply(data)
        log.info_rank(
            "  Created default noise model in", comm=data.comm.comm_world, timer=timer
        )

    # Simple elevation-weighted model
    if job_ops.elevation_model.enabled:
        job_ops.elevation_model.detector_pointing = job_ops.det_pointing_azel
        log.info_rank(
            "  Running elevation-weighted noise model...", comm=data.comm.comm_world
        )
        job_ops.elevation_model.apply(data)
        log.info_rank(
            "  Created elevation noise model in", comm=data.comm.comm_world, timer=timer
        )

    # Add common noise modes
    if job_ops.common_mode_noise.enabled:
        log.info_rank("  Running common mode noise model...", comm=data.comm.comm_world)
        job_ops.common_mode_noise.apply(data)
        log.info_rank(
            "  Added common mode noise model in", comm=data.comm.comm_world, timer=timer
        )
